import asyncio
import logging
from typing import TypeVar, Generic, Callable

T = TypeVar("T")


class TaskPool(Generic[T]):
    """
    A TaskPool is a utility that receives submitted tasks which have an identity and which should be executed no more
    than once. It can be configured to wait for the task to finish when started, wait for the task to finish when it
    is in progress, and record the result of the task when it finishes and return it when a task with the same ID is
    submitted again.
    """

    _logger = logging.getLogger(__name__)

    def __init__(self, remember_result: bool = True, wait_if_in_progress: bool = True,
                 wait_if_just_started: bool = True, restart_if_finished: bool = False):
        """
        Parameters
        ----------
        remember_result : bool
            Whether to store the result of submitted tasks

        wait_if_just_started : bool
            Whether to await a newly started task or return immediately

        wait_if_in_progress : bool
            Whether to await a task that is in progress or return immediately
        """

        if not wait_if_in_progress and remember_result:
            raise ValueError("Can't remember result without waiting")

        self._results: dict[object, asyncio.Event | T | None] = {}
        self._semaphore = asyncio.Semaphore()
        self._remember_result: bool = remember_result
        self._wait_if_in_progress: bool = wait_if_in_progress
        self._wait_if_just_started: bool = wait_if_just_started
        self._restart_if_finished: bool = restart_if_finished

    def add_values(self, results: dict[object, T]) -> None:
        """
        Adds the specified values from the dictionary as stored results.

        Parameters
        ----------
        results
            The results to store, where the key is the ID and the value is the result

        Returns
        -------
            None
        """
        self._results.update(**results)

    async def submit(self, task_id: object, func, **kwargs) -> T | None:
        """
        Submits a task and executes it. Depending on configuration, waits for it and returns the result of it.

        Parameters
        ----------
        task_id: object
            The identity of the task. E.g. "task_get_file_1" or "task_get_file_2"

        func : lambda
            The function to execute.
            Should be provided in the form of a lambda which executes some async function without using the await keyword.
            E.g::
                # Right
                taskpool.submit(..., func=lambda: some_async_function(some_val))
                # Wrong
                taskpool.submit(..., func=lambda: async some_async_function(some_val))
                #                                 ^^^^^
        kwargs
            Extra arguments for the lambda
        Returns
        -------
        T | None
            None, if the configuration states that we should not wait for the function to finish or not to remember the result.
            Returns the result of the function otherwise.
        """
        sem = self._semaphore
        await sem.acquire()

        if task_id in self._results.keys():
            if not self._wait_if_in_progress and isinstance(self._results[task_id], asyncio.Event):
                self._logger.debug("Task %s in progress but configured to not wait, returning None.", task_id)
                sem.release()
                return None

            if not isinstance(self._results[task_id], asyncio.Event):
                if self._restart_if_finished:
                    self._logger.debug("Task %s already finished but configured to restart if finished, restarting.",
                                       task_id)
                    # start_task releases the semaphore, no need to do it here
                    return await self._start_task(task_id, func, **kwargs)

                self._logger.debug("Task %s already finished, returning.", task_id)
                sem.release()
                return self._results[task_id]

            self._logger.debug("Task %s in progress. Waiting.", task_id)

            event: asyncio.Event = self._results[task_id]
            sem.release()

            await event.wait()

            self._logger.debug("Finished waiting for %s.", task_id)

            return self._results[task_id]
        else:
            # start_task releases the semaphore, no need to do it here
            return await self._start_task(task_id, func, **kwargs)

    async def _start_task(self, task_id: object, func, **kwargs):
        """
        Starts a task and **releases the results list semaphore**

        Parameters
        ----------
        task
            The task to start
        task_id
            The ID of the task
        event
            The event that the task is attached to

        Returns
        -------
        T
            The result of the task
        """
        sem = self._semaphore

        self._logger.debug("Task %s started.", task_id)

        event = asyncio.Event()
        self._results[task_id] = event
        sem.release()

        if not self._wait_if_just_started:
            # noinspection PyAsyncCall
            asyncio.create_task(self._handle_task(func, task_id, event, func_args=kwargs))
            return None

        return await self._handle_task(func, task_id, event, func_args=kwargs)

    async def _handle_task(self, func: Callable, task_id: object, event: asyncio.Event, func_args: dict) -> T:
        """
        Handles the specified task

        Parameters
        ----------
        func
            The task to handle
        task_id
            The ID of the task
        event
            The event that the task is attached to

        Returns
        -------
        T
            The result of the task
        """
        sem = self._semaphore

        result = await func(**func_args)

        if isinstance(result, asyncio.Event):
            self._logger.warning("Result was of type asyncio.Event, this will break things!")

        async with sem:
            self._logger.debug("Task %s finished", task_id)
            # Record this task as done, storing the result if configured to
            self._results[task_id] = result if self._remember_result else None
            event.set()

        return result

    async def wait_if_in_progress(self, task_id: object):
        """
        Waits for a task if it is in progress. Returns immediately otherwise.

        Parameters
        ----------
        task_id
            The task id to wait for
        """
        sem = self._semaphore

        await sem.acquire()

        if task_id in self._results and isinstance(self._results[task_id], asyncio.Event):
            event: asyncio.Event = self._results[task_id]
            sem.release()

            await event.wait()

    def clear(self) -> None:
        """
        Deletes the stored results

        Returns
        -------
        None
        """
        self._results.clear()

    def results(self) -> list[T]:
        """
        Gets the results of all currently completed tasks

        Returns
        -------
        list[T]
            The results of all currently completed tasks, excluding Nones
        """

        def filter_func(it):
            return not isinstance(it, asyncio.Event) and it is not None

        return list(filter(filter_func, self._results.values()))

    def get_completed_result_or_nothing(self, task_id: object, default: T | None = None) -> T | None:
        """
        Returns the result of an already completed task, or nothing at all if result with that id exists or is still in progress

        Returns
        -------
        object
            The result of the task

        Raises
        ------
        ValueError
            If the task is still in progress

        KeyError
            If the task has not been started yet or does not exist
        """
        if task_id in self._results:
            result = self._results[task_id]

            if isinstance(result, asyncio.Event):
                return None

            return result
        else:
            return None



from typing import Union, Optional, Dict, Any

import httpx
from gql.transport.httpx import HTTPXAsyncTransport
from graphql import DocumentNode


class CustomHTTPXAsyncTransport(HTTPXAsyncTransport):
    """:ref:`Async Transport <async_transports>` used to execute GraphQL queries
    on remote servers.

    The transport uses the httpx library with anyio.
    """

    def __init__(self, client: httpx.AsyncClient, url: Union[str, httpx.URL], headers: dict[str, Any] | None = None,
                 **kwargs):
        super().__init__(url=url, **kwargs)
        self.client = client
        self.headers = headers or {}

    async def connect(self):
        pass

    async def close(self):
        """Do not close the client. We want to keep it."""
        pass

    def _prepare_request(self, document: DocumentNode, variable_values: Optional[Dict[str, Any]] = None,
                         operation_name: Optional[str] = None, extra_args: Optional[Dict[str, Any]] = None,
                         upload_files: bool = False) -> Dict[str, Any]:
        result = super()._prepare_request(document, variable_values, operation_name, extra_args, upload_files)
        result["headers"] = self.headers

        return result



import traceback
from threading import Semaphore
from typing import Optional

from PySide6.QtCore import Slot, QUrl
from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import QDialog, QWidget, QHBoxLayout, QLabel, QLineEdit, QVBoxLayout, \
    QDialogButtonBox, QGridLayout, QMessageBox
from PySide6.QtWidgets import QProgressBar
from qasync import asyncSlot

from qcanvas.net.canvas import CanvasClient
from qcanvas.util.app_settings import settings

tutorial_url = "https://www.iorad.com/player/2053777/Canvas---How-to-generate-an-access-token-"


def row(name: str) -> QWidget:
    widget = QWidget()
    layout = QHBoxLayout()

    layout.addWidget(QLabel(name))
    layout.addWidget(QLineEdit())

    widget.setLayout(layout)

    return widget


class SetupDialog(QDialog):
    """
    The dialog shown to the user when the canvas api key/url is invalid, such as the first time the user is opening the application.
    The dialog asks for an api key and canvas url then verifies them before saving them to the primary app settings file.
    """
    def __init__(self, parent: Optional[QWidget] = None, allow_cancel: bool = True):
        super().__init__(parent)

        self.setWindowTitle("Setup")
        self._row_counter = 0
        self._operation_sem = Semaphore()
        self.allow_cancel = allow_cancel

        # Progress bar used to indicate that an operation is underway and the application has not frozen
        self.operation_activity_indicator = QProgressBar()
        self.operation_activity_indicator.setMaximum(0)
        self.operation_activity_indicator.setMinimum(0)
        self.operation_activity_indicator.setValue(0)
        self.operation_activity_indicator.hide()

        self.grid = QGridLayout()

        # Line edits for the different properties
        self.canvas_url = QLineEdit(settings.canvas_url or "")
        self.panopto_url = QLineEdit()
        self.canvas_api_key = QLineEdit(settings.api_key or "")

        # Add the line edits to the dialog
        self._row("Canvas URL", self.canvas_url)
        self._row("Painopto URL", self.panopto_url)
        self._row("Canvas API key", self.canvas_api_key)

        # Add the activity indicator to the dialog
        self.grid.addWidget(self.operation_activity_indicator, self._row_counter, 0, 1, 2)

        # Setup the rest of the layout
        grid_widget = QWidget()
        grid_widget.setLayout(self.grid)

        stack = QVBoxLayout()
        stack.addWidget(grid_widget)
        stack.addWidget(self._setup_button_box())

        self.setLayout(stack)
        self.resize(500, 200)

    def _setup_button_box(self) -> QDialogButtonBox:
        button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel if self.allow_cancel else QDialogButtonBox.StandardButton.Save)
        # Add a help button to show the user how to get a canvas api key
        button_box.addButton("How to get a canvas API key?", QDialogButtonBox.ButtonRole.HelpRole)
        # Connect signals
        button_box.helpRequested.connect(self._show_help)
        button_box.accepted.connect(self._verify)
        button_box.rejected.connect(lambda: self.reject())

        return button_box

    @asyncSlot()
    async def _verify(self) -> None:
        """
        Verifies the user's inputs before saving them
        """
        if self._operation_sem.acquire(False):
            try:
                canvas_url_text = self.ensure_protocol(self.canvas_url.text().strip())
                panopto_url_text = self.ensure_protocol(self.panopto_url.text().strip())
                canvas_api_key_text = self.canvas_api_key.text().strip()

                if not (len(canvas_url_text) > 0 and QUrl(canvas_url_text).isValid()):
                    self._show_invalid_msgbox("Canvas URL")
                    return
                if not (len(panopto_url_text) > 0 and QUrl(panopto_url_text).isValid()):
                    self._show_invalid_msgbox("Panopto URL")
                    return
                elif not len(canvas_api_key_text) > 0:
                    self._show_invalid_msgbox("API key")
                elif not (await self._verify_canvas_config(canvas_url_text, canvas_api_key_text)):
                    # Show message box saying that either the url or api key is incorrect
                    QMessageBox(
                        QMessageBox.Icon.Critical,
                        "Error",
                        f"The canvas URL or API key is invalid.\nPlease check you entered them correctly.",
                        parent=self
                    ).show()
                else:
                    # If nothing was wrong, everything should be fine
                    # Save the url and api key
                    settings.canvas_url = canvas_url_text
                    settings.panopto_url = panopto_url_text
                    settings.api_key = canvas_api_key_text

                    self.accept()
            finally:
                self._operation_sem.release()
        else:
            QMessageBox(QMessageBox.Icon.Critical, "Error", "An operation is already in progress", parent=self).show()

    def _show_invalid_msgbox(self, field_name: str) -> None:
        """
        Shows a message box saying that that specified field is invalid
        """
        QMessageBox(
            QMessageBox.Icon.Critical,
            "Error",
            f"{field_name} is invalid",
            parent=self
        ).show()

    @staticmethod
    def ensure_protocol(url: str) -> str:
        # Check if the url is blank/empty so we can tell if the user didn't input anything
        if len(url) > 0 and not (url.startswith("http://") or url.startswith("https://")):
            return "https://" + url
        else:
            return url

    async def _verify_canvas_config(self, canvas_url: str, api_key: str) -> bool:
        """
        Makes a network request to canvas to ensure the provided url and api key are correct

        Returns
        -------
        bool
            True if valid, False if invalid
        """
        self.operation_activity_indicator.show()

        try:
            return await CanvasClient.verify_config(canvas_url, api_key)
        except:
            traceback.print_exc()
            return False
        finally:
            self.operation_activity_indicator.hide()

    @Slot()
    def _show_help(self) -> None:
        msg = QMessageBox(
            QMessageBox.Icon.Information,
            "Help",
            """An interactive tutorial will open in your browser when you click OK.
            
Note that the "purpose" text doesn't matter and you can enter anything you want.

You should also leave the "expires" item blank if you want the key to last forever.

Don't share this key. You can revoke it at any time.""",
            parent=self
        )
        msg.accepted.connect(lambda: QDesktopServices.openUrl(tutorial_url))
        msg.show()

    def _row(self, name: str, widget: QWidget) -> None:
        """
        Shortcut to add a field with a label to the dialog
        """
        self.grid.addWidget(QLabel(name), self._row_counter, 0)
        self.grid.addWidget(widget, self._row_counter, 1)

        self._row_counter += 1



import asyncio
import logging
import sys

import httpx
from PySide6.QtCore import Signal
from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import QApplication, QProgressDialog, QMainWindow
from httpx import URL
from qasync import QEventLoop, asyncSlot
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio.session import async_sessionmaker as AsyncSessionMaker

import qcanvas.db as db
# noinspection PyUnresolvedReferences
import qcanvas.icons
from qcanvas.net.canvas.canvas_client import CanvasClient
from qcanvas.ui.main_ui import AppMainWindow
from qcanvas.ui.setup_dialog import SetupDialog
from qcanvas.util import self_updater
from qcanvas.util.app_settings import settings
from qcanvas.util.constants import app_name, updated_and_needs_restart_return_code
from qcanvas.util.course_indexer import DataManager
from qcanvas.util.helpers import theme_helper
from qcanvas.util.link_scanner import CanvasFileScanner
from qcanvas.util.link_scanner.canvas_media_object_scanner import CanvasMediaObjectScanner
from qcanvas.util.link_scanner.dropbox_scanner import DropboxScanner

engine = create_async_engine("sqlite+aiosqlite:///canvas_db.😘", echo=False)

logging.basicConfig()
logging.getLogger("canvas_client").setLevel(logging.DEBUG)


class LoaderWindow(QMainWindow):
    """
    Responsible for verifying that the api key and canvas url is valid, then starting the main app.
    """
    init = Signal()
    setup = Signal()
    ready = Signal()

    def __init__(self):
        super().__init__()
        self.main_window: AppMainWindow | None = None
        self.main_icon = QPixmap(":/main_icon.svg")

        self.init.connect(self.on_init)
        self.setup.connect(self.on_setup)
        self.ready.connect(self.on_ready)

        self.setWindowTitle(app_name)
        self.setWindowIcon(self.main_icon)
        self.setCentralWidget(QProgressDialog("Verifying config", None, 0, 0))

        self.init.emit()

    @asyncSlot()
    async def on_init(self) -> None:
        try:
            all_set = None not in [settings.api_key, settings.canvas_url, settings.panopto_url]
            # Verify that the canvas urls and api key are valid
            if not all_set or not await CanvasClient.verify_config(settings.canvas_url, settings.api_key):
                # Show the setup dialog
                self.setup.emit()
            else:
                # Proceed to main app
                self.ready.emit()
        except:
            # If a problem occurred then something is probably invalid
            self.setup.emit()

    @asyncSlot()
    async def on_setup(self) -> None:
        """
        Shows the setup dialog
        """
        setup = SetupDialog(self, allow_cancel=False)
        setup.setWindowIcon(self.main_icon)
        setup.rejected.connect(lambda: self.close())
        setup.accepted.connect(self.on_ready)
        setup.show()

    @asyncSlot()
    async def on_ready(self) -> None:
        """
        Sets up the canvas client and data manager for the main app
        """
        client = CanvasClient(canvas_url=URL(settings.canvas_url), api_key=settings.api_key)
        data_manager = DataManager(
            client=client,
            link_scanners=[
                CanvasFileScanner(client),
                DropboxScanner(httpx.AsyncClient()),
                CanvasMediaObjectScanner(client.client)
            ],
            # Don't expire on commit because we need to take objects outside of the session
            sessionmaker=AsyncSessionMaker(engine, expire_on_commit=False)
        )

        await data_manager.init()
        self.close()
        self.open_main_app(data_manager)

    def open_main_app(self, data_manager: DataManager) -> None:
        """
        Starts the main app
        Parameters
        ----------
        data_manager
            The data manager the app will use
        """

        self.main_window = AppMainWindow(data_manager)
        self.main_window.setWindowTitle(app_name)
        self.main_window.setWindowIcon(self.main_icon)
        self.main_window.show()
        # Set the main window as the parent of this window so this window is destroyed when the main window is closed
        self.setParent(self.main_window)


async def setup_db():
    # Create meta stuff
    async with engine.begin() as conn:
        await conn.run_sync(db.Base.metadata.create_all)


if __name__ == '__main__':
    asyncio.run(setup_db())

    app = QApplication(sys.argv)

    # Apply the selected theme to qt
    theme_helper.apply_selected_theme()

    # Setup event loop for qasync
    event_loop = QEventLoop()
    asyncio.set_event_loop(event_loop)
    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set)

    # Start the loader, which verifies the client config and initialises the data manager for the main program
    loader_window = LoaderWindow()
    loader_window.show()

    # For qasync
    with event_loop:
        event_loop.run_until_complete(app_close_event.wait())

    print("Exiting")

    # Pass the 'restart needed' flag back to the launcher script, which will re-run the program.
    # See self_updater.do_update
    if self_updater.restart_flag:
        sys.exit(updated_and_needs_restart_return_code)



import asyncio
import logging
from dataclasses import dataclass
from typing import Sequence

from bs4 import Tag, BeautifulSoup
from sqlalchemy.ext.asyncio import AsyncSession

import qcanvas.db as db
from qcanvas.util.link_scanner import ResourceScanner
from qcanvas.util.progress_reporter import ProgressReporter
from qcanvas.util.task_pool import TaskPool

_logger = logging.getLogger(__name__)

resource_elements = ["a", "iframe", "img"]


# todo could probably just use the database types directly now
@dataclass
class TransientResourceToPageLink:
    """
    Represents a temporary link between a page and a resource that will be added to the database soon.
    """
    page_id: str
    resource_id: str

    def __hash__(self):
        return hash(self.page_id) ^ hash(self.resource_id)


async def create_module_item_resource_relations(relations: Sequence[TransientResourceToPageLink],
                                                session: AsyncSession):
    """
    Creates a link between module items/pages and resources found on those pages
    """
    for relation in relations:
        existing_relation = await session.get(
            db.ResourceToModuleItemAssociation,
            (relation.page_id, relation.resource_id)
        )

        if existing_relation is None:
            session.add(
                db.ResourceToModuleItemAssociation(
                    module_item_id=relation.page_id,
                    resource_id=relation.resource_id
                )
            )


async def create_assignment_resource_relations(relations: Sequence[TransientResourceToPageLink], session: AsyncSession):
    """
    Turns temporary TransientResourceToPageLink into a persistent relation in the database
    """
    for relation in relations:
        if await session.get(db.ResourceToAssignmentAssociation, (relation.page_id, relation.resource_id)) is None:
            session.add(
                db.ResourceToAssignmentAssociation(
                    assignment_id=relation.page_id,
                    resource_id=relation.resource_id
                )
            )


# todo change resource system to think of resources as links on a page with a shallow id (that may be the same as the deep id) which links to one or more deep ids
async def find_resources_in_pages(link_scanners: Sequence[ResourceScanner], resource_pool: TaskPool[db.Resource],
                                  items: Sequence[db.PageLike], progress_reporter: ProgressReporter) -> list[
    TransientResourceToPageLink]:
    """
    Produce a list of resource to page links from resources extracted from the specified `items` using `link_scanners`.
    Extracted resources will be added to `resource_pool`
    """
    progress = progress_reporter.section("Indexing resources", len(items))
    tasks = []

    for item in items:
        # Assignment descriptions may be null. Avoid creating extra tasks by checking here
        if item.content is None:
            progress.increment_progress()
            continue

        # extract resources from the page
        task = asyncio.create_task(_extract_resources_from_page(link_scanners, resource_pool, item))
        task.add_done_callback(progress.increment_progress)
        tasks.append(task)

    if len(tasks) > 0:
        # Wait for all tasks to complete
        await asyncio.wait(tasks)

        result = []
        # Flatten the array of results
        for task in tasks:
            result.extend(task.result())

        return result
    else:
        return []


async def _extract_resources_from_page(link_scanners: Sequence[ResourceScanner], resource_pool: TaskPool[db.Resource],
                                       page: db.PageLike) -> list[TransientResourceToPageLink]:
    """
    Extracts any detected resource links from the specified page and then uses `link_scanners` to extract information
    about which is then added to the `resource_pool`.

    Returns
    -------
    list
        A list of resource to page links for any resources found on this page.
    """
    _logger.debug(f"Scanning %s %s for files", page.id, page.name)
    tasks = []

    # Extract iframes, hyperlinks, etc from the page
    for link in _scan_page_for_links(page):
        tasks.append(asyncio.create_task(_process_link(link_scanners, resource_pool, link, page.course_id)))

    if len(tasks) > 0:
        # Wait for all tasks to complete
        await asyncio.wait(tasks)

        task_results = [task.result() for task in tasks]

        # Convert every non-null result in the task results to a resource page link and return it
        return [
            TransientResourceToPageLink(page_id=page.id, resource_id=result.id)
            for result in task_results if result is not None
        ]
    else:
        return []


def _scan_page_for_links(page: db.PageLike) -> list[Tag]:
    """
    Extracts (potential) resource elements from a PageLike object
    """
    soup = BeautifulSoup(page.content, 'html.parser')
    return list(soup.find_all(resource_elements))


async def _process_link(link_scanners: Sequence[ResourceScanner], resource_pool: TaskPool[db.Resource], link: Tag,
                        course_id: str) -> db.Resource | None:
    """
    Iterates over `link_scanners` to find one that will accept `link`, then uses it to fetch resource information and
     adds it to the `resource_pool`.
    If no scanner accepts the link then None is returned.
    """
    for scanner in link_scanners:
        if scanner.accepts_link(link):
            resource_id = scanner.extract_id(link)

            return await resource_pool.submit(
                f"{scanner.name}:{resource_id}",  # match the format used by the resource id
                lambda: _extract_file_info(link, scanner, resource_id, course_id)
            )

    return None


async def _extract_file_info(link: Tag, scanner: ResourceScanner, resource_id: str,
                             course_id: str) -> db.Resource | None:
    """
    Extracts file info from `link` using `scanner` and assigns the course_id to the resulting resource.

    Parameters
    ----------
    link
        The html element to scan
    scanner
        The scanner to process the link with
    course_id
        The id of the course the file belongs to
    Returns
    -------
    db.Resource
        The resource if the link was processed successfully.
    None
        If processing failed
    """
    try:
        _logger.debug(f"Fetching info for file %s with scanner %s", scanner.extract_id(link), scanner.name)

        result = await scanner.extract_resource(link, resource_id)
        result.id = f"{scanner.name}:{result.id}"  # Prefix the scanner name to prevent resources from different sites potentially clashing
        result.course_id = course_id
        return result
    except BaseException as e:
        _logger.error(f"Failed to retrieve info for file id %s: %s", f"{scanner.name}:{resource_id}", str(e))
        return None



import asyncio
from typing import Callable, Any

from PySide6.QtCore import QObject, Signal

from qcanvas.util.task_pool import TaskPool


class DownloadPool(TaskPool[None], QObject):
    """
    Like the TaskPool, but uses a queue (a channel in kotlin terms) in the task to report download progress.
    When using .submit(), it is EXPECTED that the lambda returns a function that uses yield to signal progress, e.g:
    ```
    pool.submit(my_task_id, lambda: my_func())

    async def my_func(channel: asyncio.Queue):
        for i in range(0, 10):
            await slow_thing()
            yield (i / 10) # can be whatever format/type you want
    ```
    """
    download_progress_updated = Signal(object, Any)
    download_failed = Signal(object)
    download_finished = Signal(object)

    def __init__(self, parent: QObject | None = None):
        QObject.__init__(self, parent)
        TaskPool.__init__(self)

    async def _handle_task(self, func: Callable, task_id: object, event: asyncio.Event, func_args: dict):
        sem = self._semaphore

        try:
            # Consume progress updates 'yield'ed from the function
            async for progress in func(**func_args):
                # Fire the download progress update signal
                self.download_progress_updated.emit(task_id, progress)
        except BaseException as e:
            # Try to maintain integrity when a task fails
            async with sem:
                # Release anything else waiting for this task
                event.set()
                # Remove the record from the results map
                del self._results[task_id]

            # Emit failure signal and rethrow
            self.download_failed.emit(task_id)
            raise e

        async with sem:
            self._logger.debug("Task %s finished.", task_id)
            # Record this task as done
            self._results[task_id] = None
            event.set()

        self.download_finished.emit(task_id)

        return None


from typing import Any

from PySide6.QtWidgets import QStatusBar, QProgressBar

from qcanvas.util.progress_reporter import ProgressReporter


class StatusBarReporter(ProgressReporter):
    def __init__(self, status_bar: QStatusBar):
        self.status_bar = status_bar
        self.section_name: None | str = None
        self.progress_bar: QProgressBar | None = None

    def section_started(self, section_name: str, total_progress: int) -> None:
        self.section_name = section_name
        self.status_bar.showMessage(section_name)

        if self.progress_bar is None:
            self.progress_bar = QProgressBar(self.status_bar)
            self.progress_bar.setMaximumHeight(self.status_bar.height())
            self.progress_bar.setMinimum(0)
            self.status_bar.addPermanentWidget(self.progress_bar)

        self.progress_bar.setValue(0)
        self.progress_bar.setMaximum(total_progress)

    def progress(self, current_progress: int, total: int) -> None:
        if self.progress_bar is not None:
            self.progress_bar.setValue(current_progress)

    def finished(self) -> None:
        self.status_bar.removeWidget(self.progress_bar)
        self.progress_bar = None
        self.status_bar.showMessage("Finished", 5000)

    def errored(self, context: Any) -> None:
        if self.status_bar.parent() is not None:
            self.status_bar.removeWidget(self.progress_bar)
            self.progress_bar = None
            self.status_bar.showMessage("Synchronisation error!!", 5000)



from .all_courses import AllCoursesQueryData
from .canvas_course_data import Term, Assignment, AssignmentConnection, ModuleItemInterface, File, Page, \
    ModuleItem, \
    Module, ModuleConnection, CanvasCourseData as Course



"""
I dedicate this file to removing the random crap canvas puts in its data.
Like random NBSPs. Gotta love those fuckers.
Thanks instructure btw for you state of the art WYSIWYG dogshit editor.
"""
from bs4 import BeautifulSoup

NBSP = " "


def remove_garbage_from_title(smelly_canvas_title: str) -> str:
    """
    Removes trailing tabs, spaces and NBSPs from smelly canvas titles.
    Parameters
    ----------
    smelly_canvas_title

    Returns
    -------
    str
        Clean title that is not smelly and has no NBSPs.
    """
    return (smelly_canvas_title
            .strip(f"\t {NBSP}")  # remove trailing garbage
            .replace(NBSP, " ")  # remove any other NBSPs
            )


def remove_stylesheets_from_html(smelly_html: str) -> str:
    """
    Removes all stylesheet links from `smelly_html`.

    Parameters
    ----------
    smelly_html
        The html to remove style sheets from
    Returns
    -------
        The non-smelly html with all stylesheet links removed
    """
    bs = BeautifulSoup(smelly_html, "html.parser")

    # remove links
    for ele in bs.find_all("link", {"rel": "stylesheet"}):
        ele.decompose()

    return str(bs)



from PySide6.QtGui import QStandardItem

from qcanvas.util import tree_util as tree


class ContainerItem(QStandardItem):
    def __init__(self, data: tree.HasText):
        super().__init__()
        self.content = data
        self.setEditable(False)
        self.setText(data.text)



from typing import Any, TypeVar, Sequence, Optional

from PySide6.QtCore import QModelIndex, Slot, QItemSelectionModel
from PySide6.QtWidgets import QTreeView, QWidget

from .model_helpers import HasChildren, HasParent
from .tree_model import TreeModel

T = TypeVar("T")


class ExpandingTreeView(QTreeView):
    """
    Provides a way to retain the collapsed or expanded state of tree items after a model reset. Expects the tree's
    model to implement `AbstractItemModelHasListRoot` and any children to implement `HasChildren`. Also provides
    some methods that can be used to get a model index for an arbitrary object in a tree. These will NOT work when
    duplicate objects are present in the tree. Does NOT support multi-selection.
    """

    def __init__(self, parent: Optional[QWidget] = None):
        """
        Constructor
        :param session_factory: The session that will be used to update an item's expanded/collapsed state when it is changed
        :param parent: The parent of this treeview
        """
        super().__init__(parent)
        self._connect_expanded_listeners()

    def reexpand(self) -> None:
        """
        Re-expands all items in the tree that have children based on their saved collapsed/expanded state.
        """
        self._disconnect_expanded_listeners()

        try:
            root: Sequence[Any] = self.model().get_root()

            # Go through every item in the root and reexpand it and any children it has
            for index, item in enumerate(root):
                if isinstance(item, HasChildren):
                    self._reexpand_recur(item, self.model().index(index, 0, QModelIndex()))
        finally:
            self._connect_expanded_listeners()

    def _connect_expanded_listeners(self):
        self.expanded.connect(self.tree_item_expanded)
        self.collapsed.connect(self.tree_item_collapsed)

    def _disconnect_expanded_listeners(self):
        self.expanded.disconnect(self.tree_item_expanded)
        self.collapsed.disconnect(self.tree_item_collapsed)

    def _reexpand_recur(self, item: HasChildren, index: QModelIndex) -> None:
        """
        Internal function to expand an item in the tree. Recurs to any children it has.
        :param item: The object to expand in the tree
        :param index: The index of that object
        """
        # Sanity check
        if not isinstance(item, HasChildren):
            raise TypeError("item must be an instance of HasChildren")

        # Expand this item
        self.setExpanded(index, not item.collapsed)

        for child_index, child in enumerate(item.get_children()):
            # Check if the child item can also have children and be collapsed. Also check that it actually has
            # any children.
            if isinstance(child, HasChildren) and len(child.get_children()) > 0:
                # Expand it and any children it has
                self._reexpand_recur(child, self.model().index(child_index, 0, index))

    @Slot()
    def tree_item_expanded(self, index: QModelIndex) -> None:
        """
        Slot that is connected to the tree's `expanded` signal
        :param index: The index that has been expanded
        """
        item = self.model().get_item(index)

        # Update the item's state
        if isinstance(item, HasChildren):
            item.collapsed = False

    @Slot()
    def tree_item_collapsed(self, index: QModelIndex) -> None:
        """
        Slot that is connected to the tree's `collapsed` signal
        :param index: The index that was collapsed
        """
        item = self.model().get_item(index)

        # Update the item's state
        if isinstance(item, HasChildren):
            item.collapsed = True

    def get_path_of_indexes_for_item(self, item: Any) -> list[int]:
        """
        Gets the path of numerical indexes for an arbitrary object in a tree.
        For example, second child of the first item would be [0, 1].

        :param item: The item to find the path for. The item must belong to the tree
        :return: The path to the item as a list of indexes to follow
        """
        path: list[int] = []

        # Keep going until we reach the root
        while isinstance(item, HasParent) and item.parent is not None:
            # Add the index of the item to the path and then go to the parent of that child until there is no parent
            path.insert(0, item.index_of_self)
            item = item.parent

        # Insert the index of the item belonging to the root list to the path
        path.insert(0, self.model().get_root().index(item))

        return path

    def select_object(self, item: Any) -> None:
        """
        Selects an arbitrary item that belongs to the tree. Will not work for branches that have duplicate items.
        :param item: The item to select. Must belong to the tree
        """
        self.select_item_by_path(self.get_path_of_indexes_for_item(item))

    def select_item_by_path(self, item_path: list) -> None:
        """
        Selects an item based on its path
        :param item_path: The path of the item
        """
        selection_model = self.selectionModel()

        # Convert the path to a model index
        # Clear previous selection and select the whole row
        selection_model.select(self.get_model_index_for_path(item_path),
                               QItemSelectionModel.SelectionFlag.ClearAndSelect | QItemSelectionModel.SelectionFlag.Rows)

    def get_model_index_for_path(self, item_path: list[int]) -> QModelIndex:
        """
        Converts an item's path into a QModelIndex
        :param item_path: The path of the item to convert
        :return: The path as a QModelIndex
        """
        model: TreeModel[Any] = self.model()
        # Get the first index
        item_index = item_path.pop(0)
        # Create the base QModelIndex
        model_index = model.index(item_index, 0, QModelIndex())
        # Get the children of the first item
        children = model.get_root()[item_index]

        while len(item_path) > 0:
            # Get the next item
            item_index = item_path.pop(0)
            # Update the QModelIndex
            model_index = model.index(item_index, 0, model_index)

            if len(item_path) > 0:
                # Sanity check
                if not isinstance(children, HasChildren):
                    raise TypeError("Parent object does not have any children but a child was expected")

                # Go to the children of this child next
                children = children.children[item_index]

        return model_index



from typing import Sequence

from PySide6.QtCore import QObject, Slot, Qt
from PySide6.QtWidgets import QStyledItemDelegate, QStyleOptionProgressBar, QApplication, QStyle, \
    QHeaderView, QStyleOptionViewItem, QTreeWidget, QTreeWidgetItem

import qcanvas.db.database as db
from qcanvas.util.download_pool import DownloadPool
from qcanvas.util.helpers import file_icon_helper


# https://code.whatever.social/questions/1094841/get-human-readable-version-of-file-size#1094933
def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


class FileColumnDelegate(QStyledItemDelegate):
    def __init__(self, tree: QTreeWidget, download_pool: DownloadPool, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.download_pool = download_pool
        self.tree = tree

    def paint(self, painter, option, index) -> None:
        item = self.tree.itemFromIndex(index)

        # Check that the item is actually a file row and not already downloaded
        if not isinstance(item, FileRow) or item.resource.state == db.ResourceState.DOWNLOADED:
            return super().paint(painter, option, index)

        resource = item.resource
        progress = item.download_progres

        if progress == resource.file_size:
            return super().paint(painter, option, index)

        # Check that the file is not currently downloading (signaled by progress = -1)
        if progress < 0:
            # Show the state of the file instead (downloaded, failed, not downloaded)
            view_item = QStyleOptionViewItem(option)
            self.initStyleOption(view_item, index)
            view_item.text = db.ResourceState.human_readable(resource.state)

            return QApplication.style().drawControl(QStyle.CE_ItemViewItem, view_item, painter, view_item.widget)
        else:
            # otherwise show progress bar
            progress_bar = QStyleOptionProgressBar()
            progress_bar.rect = option.rect
            progress_bar.minimum = 0
            progress_bar.maximum = resource.file_size
            progress_bar.progress = progress
            # Display download progress as a percentage
            progress_bar.text = "{0:.0f}%".format((progress / resource.file_size) * 100)
            progress_bar.textVisible = True

            QApplication.style().drawControl(QStyle.CE_ProgressBar, progress_bar, painter)


class FileRow(QTreeWidgetItem):
    def __init__(self, resource: db.Resource):
        # Set column names
        super().__init__(
            [
                resource.file_name,
                resource.date_discovered.strftime("%Y-%m-%d"),
                sizeof_fmt(resource.file_size),
            ]
        )

        # Set the icon for this file, displayed in the name column
        self.setIcon(0, file_icon_helper.icon_for_filename(resource.file_name))

        self.resource = resource
        # This is used to pass the download progress value from download_progress_updated to FileColumnDelegate. This way
        # it means we don't have to get the download progress from the DownloadPool a second time with an async function
        # (which is undesirable from an item delegate...). Use -1 to show that the file is not currently downloading.
        self.download_progres = -1


class FileList(QTreeWidget):
    def __init__(self, download_pool: DownloadPool, parent: QObject | None = None):
        super().__init__(parent)
        self._setup_header()

        # Keep a map of file ids to rows, so we can actually update the download column using a file id.
        # This has to be a list or else files that are in multiple modules/pages will not be updated properly on the tree
        self.row_id_map: dict[str, list[FileRow]] = {}
        self.download_pool = download_pool
        self.setAlternatingRowColors(True)
        # Add the download column renderer
        self.setItemDelegateForColumn(3, FileColumnDelegate(tree=self, download_pool=download_pool, parent=None))

        # Connect download update events
        self.download_pool.download_progress_updated.connect(self._file_download_progress_update)
        self.download_pool.download_failed.connect(self._file_download_failed_update)

    @Slot(str, int)
    def _file_download_progress_update(self, resource_id: str, progress: int):
        # Ignore files that aren't in this tree
        if resource_id not in self.row_id_map:
            return

        rows = self.row_id_map[resource_id]

        # Update each row that maps to this file.
        # Note that a file may belong to multiple modules or pages
        for row in rows:
            # Update the progress
            row.download_progres = progress

            # Get the download column for the relevant item
            index = self.indexFromItem(row, 3)

            # Update the download column
            self.model().dataChanged.emit(index, index, Qt.ItemDataRole.DisplayRole)

    @Slot(str)
    def _file_download_failed_update(self, resource_id: str):
        if resource_id not in self.row_id_map:
            return

        rows = self.row_id_map[resource_id]

        for row in rows:
            # Get the download column for the relevant item
            index = self.indexFromItem(row, 3)

            # Update the download column
            self.model().dataChanged.emit(index, index, Qt.ItemDataRole.DisplayRole)

    def _setup_header(self):
        self.setColumnCount(4)
        self.setHeaderLabels(["Name", "Date Found", "Size", "Download"])

        header = self.header()
        header.setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(2, QHeaderView.ResizeMode.ResizeToContents)
        # Make the download progress column not try to take up all remaining space in the tree, as it just looks weird
        # and decreases room for more important data
        header.setStretchLastSection(0)

    def load_items(self, items: Sequence[db.ModuleItem | db.Module]):
        self.clear()

        groups = []

        for item in items:
            # Get the resources depending on the type of the item
            if isinstance(item, (db.PageLike, db.ModuleItem)):
                resources = item.resources
            elif isinstance(item, db.Module):
                resources = []

                for module_item in item.items:
                    resources.extend(module_item.resources)
            else:
                continue

            # Avoid adding groups with no files in them, it just adds clutter
            if len(resources) == 0:
                continue

            # Create the group node for this item
            group_node = QTreeWidgetItem([item.name])

            # fixme this does not remove duplicate files e.g. when on module groups

            for resource in resources:
                row = FileRow(resource)

                # Put the file into the row id map so we can use the file id provided by download_progress_updated
                # and such to update the download progress in the tree
                if resource.id not in self.row_id_map:
                    self.row_id_map[resource.id] = [row]
                else:
                    self.row_id_map[resource.id].append(row)

                # Add the row to the group
                group_node.addChild(row)

            # Add the module/page to the list
            groups.append(group_node)

        # Add all the items and expand them
        self.insertTopLevelItems(0, groups)
        self.expandAll()

    def clear(self):
        super().clear()
        self._setup_header()
        self.row_id_map.clear()



from abc import abstractmethod
from typing import Sequence

from PySide6.QtCore import QItemSelection, Slot
from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QWidget, QTextBrowser, QTreeView, QHBoxLayout
from bs4 import BeautifulSoup

import qcanvas.db as db
from qcanvas.ui.container_item import ContainerItem
from qcanvas.util.constants import default_assignments_module_names
from qcanvas.util.course_indexer import resource_helpers
from qcanvas.util.helpers import canvas_sanitiser
from qcanvas.util.link_scanner import ResourceScanner


class LinkTransformer:
    # This is used to indicate that a "link" is actually a resource. The resource id is concatenated to this string.
    # It just has to be a valid url or qt does not send it to anchorClicked properly
    transformed_url_prefix = "data:,"

    def __init__(self, link_scanners: Sequence[ResourceScanner], files: dict[str, db.Resource]):
        self.link_scanners = link_scanners
        self.files = files

    def transform_links(self, html: str):
        doc = BeautifulSoup(html, 'html.parser')

        for element in doc.find_all(resource_helpers.resource_elements):
            for scanner in self.link_scanners:
                if scanner.accepts_link(element):
                    resource_id = f"{scanner.name}:{scanner.extract_id(element)}"
                    # todo make images actually show on the viewer page if they're downloaded
                    if resource_id in self.files:
                        file = self.files[resource_id]

                        substitute = doc.new_tag(name="a")
                        # Put the file id on the end of the url so we don't have to use the scanners to extract an id again..
                        # The actual url doesn't matter
                        substitute.attrs["href"] = f"{self.transformed_url_prefix}{file.id}"
                        substitute.string = f"{file.file_name} ({db.ResourceState.human_readable(file.state)})"

                        element.replace_with(substitute)
                    else:
                        if element.string is not None:
                            element.string += " (Failed to index)"

                    break

        return str(doc)


class PageLikeViewer(QWidget):
    def __init__(self, header_name: str, link_transformer: LinkTransformer):
        super().__init__()
        self.viewer = QTextBrowser()
        self.viewer.setOpenLinks(False)

        # todo just use QTreeWidget instead
        self.tree = QTreeView()
        self.model = QStandardItemModel()
        self.header_name = header_name
        self.link_transformer = link_transformer

        self.tree.setModel(self.model)
        self.tree.selectionModel().selectionChanged.connect(self.on_item_clicked)
        self.model.setHorizontalHeaderLabels([self.header_name])

        layout = QHBoxLayout()
        layout.addWidget(self.tree)
        layout.addWidget(self.viewer)
        layout.setStretch(1, 1)
        self.setLayout(layout)

    def fill_tree(self, data: db.Course):
        self.model.clear()
        self.viewer.clear()
        self._internal_fill_tree(data)
        self.model.setHorizontalHeaderLabels([self.header_name])
        self.tree.expandAll()

    def clear(self):
        self.model.clear()
        self.viewer.clear()
        self.model.setHorizontalHeaderLabels([self.header_name])

    @abstractmethod
    def _internal_fill_tree(self, data: db.Course):
        ...

    @Slot(QItemSelection, QItemSelection)
    def on_item_clicked(self, selected: QItemSelection, deselected: QItemSelection):
        if len(self.tree.selectedIndexes()) == 0:
            return

        node = self.model.itemFromIndex(self.tree.selectedIndexes()[0])

        if isinstance(node, ContainerItem):
            item = node.content

            if isinstance(item, db.PageLike):
                if item.content is None:
                    return

                # todo when a file is finished downloading it would be nice if the page was refreshed to show the state properly
                html = canvas_sanitiser.remove_stylesheets_from_html(item.content)
                self.viewer.setHtml(self.link_transformer.transform_links(html))


class PagesViewer(PageLikeViewer):
    def __init__(self, link_transformer: LinkTransformer):
        super().__init__("Pages", link_transformer)

    def _internal_fill_tree(self, course: db.Course):
        root = self.model.invisibleRootItem()

        for module in course.modules:
            if module.name.lower() in default_assignments_module_names:
                continue

            module_node = ContainerItem(module)
            module_node.setSelectable(False)

            for module_item in list[db.ModuleItem](module.items):
                module_node.appendRow(ContainerItem(module_item))

            root.appendRow(module_node)


class AssignmentsViewer(PageLikeViewer):

    def __init__(self, link_transformer: LinkTransformer):
        super().__init__("Putting the ASS in assignments", link_transformer)

    def _internal_fill_tree(self, course: db.Course):
        root = self.model.invisibleRootItem()

        default_assessments_module = None

        for module in course.modules:
            if module.name.lower() in default_assignments_module_names:
                default_assessments_module = module
                break

        if default_assessments_module is not None:
            for module_item in default_assessments_module.items:
                root.appendRow(ContainerItem(module_item))

        for assignment in course.assignments:
            root.appendRow(ContainerItem(assignment))



from typing import Any

from PySide6.QtGui import QAction, QKeySequence


def create_qaction(name: str, shortcut: QKeySequence | None = None, parent: Any = None, triggered: Any = None,
                   checkable: bool | None = None, checked: bool | None = None) -> QAction:
    action = QAction(name)

    if shortcut is not None:
        action.setShortcut(shortcut)

    if parent is not None:
        action.setParent(parent)

    if triggered is not None:
        action.triggered.connect(triggered)

    if checkable is not None:
        action.setCheckable(checkable)

        if checked is not None:
            action.setChecked(checked)

    return action



import httpx
from bs4 import Tag
from httpx import URL

from qcanvas import db as db
from qcanvas.util.link_scanner import ResourceScanner


# from httpx import URL


def parse_content_disposition(header: str) -> dict[str, str | None]:
    bad_chars = "\" \t"
    result = {}

    for statement in header.split(";"):
        split = statement.split("=", 2)

        result[split[0].strip(bad_chars)] = None if len(split) == 1 else split[1].strip(bad_chars)

    return result


class DropboxScanner(ResourceScanner):
    def __init__(self, client: httpx.AsyncClient):
        self.client = client

    def accepts_link(self, link: Tag) -> bool:
        if link.name != "a":
            return False

        if "href" in link.attrs:
            url = URL(link.attrs["href"])

            return url.host == "www.dropbox.com" and url.path.split("/", 2)[1] == "s"
        else:
            return False

    async def extract_resource(self, link: Tag, file_id: str) -> db.Resource:
        url = URL(link.attrs["href"]).copy_set_param("dl", 1)

        req = self.client.build_request(
            method="GET",
            url=url
        )

        # The following code essentially starts downloading the file, reads the headers and then stops downloading it,
        # just to ge the size of the file
        resp = await self.client.send(req, follow_redirects=True, stream=True)

        try:
            resp.raise_for_status()

            filename = parse_content_disposition(resp.headers["content-disposition"])["filename"]
            size = int(resp.headers["content-length"])

            return db.Resource(id=file_id, url=str(url), file_name=filename, file_size=size)
        finally:
            await resp.aclose()

    def extract_id(self, link: Tag) -> str:
        # https://www.dropbox.com/s/vwk48ajl9nw6pqh/Lab1_ENGG1500_robot.pdf?dl=0
        # ------- Extract this part ^^^^^^^^^^^^^^^
        return URL(link.attrs["href"]).path.split("/", 3)[2]

    @property
    def name(self) -> str:
        return "dropbox"



# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 6.6.1
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore

qt_resource_data = b"\
\x00\x00\x0c\xda\
\x00\
\x00:\xf7x\xda\xdd[\xeb\x8f\xdb6\x12\xff\xbe\x7f\x85\
\xce\xc1\xa1\x09\x22\xc9\x22\xa9\xa7\xf7Q4\x09z\x09p\
\x87\x03.i\x0f\xe8\x97B\x96h[]=\x0cI^\
\xdb9\xdc\xff~3\xd4[\xa6\xed\x8d\xbb\xdd\x5c\xbaF\
di8\x1c\x0e\x7f3C\xceP\xce\xcd\xf7\xbb$V\
\x1ex^DYz;!\xba1Qx\x1ada\x94\
.o'?}\xfaQs'JQ\xfai\xe8\xc7Y\
\xcao'i6\xf9\xfe\xee\xea\xe6/\x9a\xa6\xbc\xcd\xb9\
_\xf2P\xd9F\xe5J\xf9\x90\xde\x17\x81\xbf\xe6\xca\xcb\
UY\xaeg\xd3\xe9v\xbb\xd5\xa3\x9a\xa8g\xf9r\xfa\
J\xd1\xb4\xbb\xab\xab\x9b\xe2ay\xa5(J\xd36\xe3\
\xbbu\x96\x97\xda>\x5cG\xb7\x13\xcf\xd0\x0d\xf17\x91\
\xf1\xec$<\xdb(,W\xb7\x13j\xd9\xe2q\xc5\xa3\
\xe5\xaa\xec\x9e\xa3\xf0v\x02#\x12\xc2j\xfe\x02\xe6\xb6\
\x86\x7f\xb3v\xd2\x86\xce\xe8p\xb8\x1e\x1eL\xa7\xcaK\
2\x137\xaf\xa9A\x19!\xd4\xa2\xc42^\x1b\x1e\xe1\
\xd4\xe0\x0bc\xf1j(8\xcc\x82\xd4O\x00\xab(\xc8\
R\x1d\x06\x1f\x0a\xcf6\xe5zS\xfe\xcaw%O\xab\
Q\x00\x9d\x1eT\xa2\x19\xbb\xb54\xd1\x7f`#1^\
\xb9\x8fa\x900*\xd6\xb1\xbf\x9fEi\x1c\xa5\xfc\x9a\
\xa7\xfe<\xe6\xda\xdc\x0f\xee\x97y\xb6I\xc3Y\xca\xb7\
R0\x17Q\xcc+=\xa7\xab,\xe1\xd3\xc2O\xa6I\
v\xefk)\xa8\xa6\x95<\x01\xb1%\xd7\xd7i\xa5?\
\xf8\xc9\xacX\xfb\x01\xf0\xafs^\xf0\xfc\x817\xf4\xb4\
\x985\xc2o'\xc7\xcc\x8fC\x89\xfe\xc5t0\xaf\xaa\
\x7f\x03^\xdb\xbf!\xe8E\xb6\xc9\x03\xbe\x00\x11\x5cO\
y9}\xf7\xe9]\xdb\xa8\x19zX\x86=1;\xc0\
\xe0~\xa0\xc3\x96\x89\xd1\x89\xe7yS\xd1\xdaq\xcb\xf8\
(\xb8\xd5\xb41Y\xad\xda\xc3\xf2\x91\x9cy\xb88:\
\xb6A\xa7\x94j\xc0\xa1\x15\xfb\xb4\xf4wZZ\xbc\xe8\
u\x0d\x82\xb6g\x80a\x15=\xf0 K\x92,-*\
\xf0\x06\xcca\xc7\xbc\xde\xe4\xb1\xe0\x08\x83)\x8fy\xc2\
\xd3\xb2\x98\x12\x9dL'w7\xadC\x22\xf4\xe1C\xc4\
\xb7(\x02\x1d'\xcf\xee\xc1R/\xf8\x82z\xd4\x9bT\
Tp\x87\x18h\x0b\xcb\xf1\xeaH\xa9\x82g\xee\x17\xbc\
~\x5c\xfbK\xd0*\xcer\xe0s\xc4_\xdd0\xcf\xf2\
\x90\xe7M\x93-\xfe\x06M\x19\x18>*\xf7\x18k\xd4\
2=\x83xM{\xeb\x95(\xbcc\x93\xb5\x16+?\
\xcc\xb6\x10\xda\xe3\xc6\xcfY\x96H\xc8\xc1\x0e\x88\xa6\xad\
;\xd6A\x0b\x0c\x01q,k\xd9\xe49`\xa8AD\
q\x98\x8b\xf8jfY\xac\xb2\xed2GL\x16~\xdc\
\x82\xd2v\xc5&m>\xcf`\xd42\xdf\x1c4\xc3\xaa\
\xb0A\xf3h\x9b4*\xc1\xf7\xd6\xbb1\x07\xca\xef\xcf\
S:\xca6J\xa1Q\xabW=\xe2Q\xe3\x08G\xb3\
\x10\x12\x83\xb8GXv'\x05\xf4\xccP\x8ff\x1aF\
\xabu#\x9e\xf5h\xddLR\x7f\xad\xa5Y\xc8\x0b9\
\x16\xa2\xfd\x00*\x04\xb0\xccb\x9e\xfb)\xae2\xa4]\
\xe2\xfb\x8b\xe7\xfc7\x1e\x94ge\xa3\xa8a\xbb0\xde\
&:\xdaM\xb4\x9d0_\x0dJ\xe2\xef\xa2$\xfa\xcc\
\xc1\x09\xc8\x98\x05;\x9f\xd2L\xb4\xaf\xfdru\x0e\x15\
\x8d\x87K\xae%Q\xb8\xce\xa2\xb4<\xcb}\x86QD\
\xed\x8a\x07\xf7<\x9fg~\x1eJ\xa2\x0bT\xbeo\xa2\
7$\xf8\x81\xe5c\xe0\xd8U\x07\x80\x11#4]b\
\xfc\xb4fo\xa9\xbb!\x15#\x05\xfbZ\xae\xcbZb\
\xb9\xc7Mb\xb7\xc7\x86\x96XmZ\xe1@\x7f\xd8\xf1\
\xa2\x22\x9a\xe3\x16\xd7\x0f\x04`N\xd6\xf5x\xe0\x91\x9d\
\x0a\x00G\xdd\x01E\xe3^Xdi\xbc\x1f\xc9\x0c\xb3\
\xb2\xe4\xe1Xd3\xf7^Z\x81\x7f\xbdU\x8b\xe0\xb2\
\xd5W\xa1\xe9B\x0d\xc6\x17\xac\xdf\xd2\xeb\xc5l\xd31\
`\xfa\x9d\xc4<ZF)\xe0\xa4\x11{D\x84\x0e\x1a\
\xed8\xbbUB\x99\x1e\xb1\x85\x14\xca\x06tB,\x8f\
J1s\xcf\x02\xdcXcH\xee\x8cl\xe8\xd6\xa1\xf1\
\xf7cz\x83\xd0b\xf1\xc5\xa0\x8e\xba\x0cA\xa5\x96\xe1\
\xb9\x06\xf3.5\xfd\x80x\xa9A\xa6\x87\x9b\xeb\xddM\
\xc8\x17E\xb7q\xe2\x13\x838B}\xfc\xfco\xb9\x1f\
F\xb0\xfc\xb7vjw\x9c,\x8eaI\xbb\x9d\xf8\xf1\
\xd6\xdf\x17\xedx\x22Y\x99\xadr\x0e9\xc5\x0bH\x0a\
y\xfe\xa6\xcd\xe9\x1aa\x1a\x1d\xd8}8\x92\x09\xe8v\
\xd2\x08,\xd5V\xfb\xb8\x87Gbu3\xdeQ\xcc\x98\
;7\xde\xd3a\xfb\xb2\x96\xf9S\x05\xc2\x06\xd2\xbf\x8f\
\x98\xcb\xfd3\xfd\xa9\xe7!\x0d\xd7'X\xc1\x0b\xc8\xd8\
`S\xce\xb3\x12\x92\xc8\x97\x9ag\xa8\xc44U\x90\xf9\
J\xe0w\x04\x14\x98\xc4\xf1\xa9BFSf\xeb\x86\xb7\
M\x80\x91\xa8\x09\xcf\x99\xbd\xe0\x8e\xe1\x1b\xfe\xb5 \xd5\
.3#\xd7\x93\xaeO\xb6X\x14\xbc\xec\x96\xc0fT\
\xec\xc1\x5c\xdb\xd4\x5c\xcd\xd6\xc0\xf5P\xc7\xb3\xa3-8\
\xba\xe9#F#\xf2\xd1l\xcd\xd3\x88F*\x87\x1a\x22\
\x82\xa3\xc3xW\x83N@\x80\x85Y\x0f\xe2B#\xff\
\xc1|\x0dTX,\xae\xff\x0b\xde\x88m \x04}\xee\
\xee&\xe1\xa5\x1f\xfa\xa5\xdf\xf9bC1\x01D\xc8A\
g\xffz\xf7\xe3\xddM\x10\xcc\xfe\x9d\xe5\xf7\x9dj\xd8\
\xe2\xcf\xc1\x00\xb7\x13\xe0\x0b\x83\x19\xda\xd0/\xef\xa2\x04\
\xb6\x10\xccu_C\xfe\x09\x83\xb4\x0d\xc8\x83\xabP'\
\xa2\x12\x02\xf5\x81\xc8\xda\xa5ij\x18$\x11v\x9a~\
,a\x0a\x1fP\xb6@\x00d\x89\xd47\xcb\x85j?\
,\x05\x0c8BT\xc2\xe4>\xfa\x89\xf2\x9eC\xadY\
\x0a\x0d*\xe2\xcd\xb4\xe3\x9c\xf6\x05\xc0}\xa5\x02J\x06\
\x9e8\x0a\xa0\xda:\xa1h\xa3\x82\x90[\xdf\x17\x1b\x91\
mT\x90\xbd\xf1\x97\xc2N}:\xdc\x03\xaa\x0d{\x8e\
\x19Q!S^t\x1cj\xda0\xc3\xedz3\x8f\xa3\
b\xc5\xf3Gv\xed\xf1\xc3\x13\xa4-i\x19-\x22\x9e\
7jp\xa8\xdb\xa0R\xac\x1fc?]n\x00\xe3\x06\
\xe2\x0c\x0a\xc9\xee\x11v\xfe \x8f\xd6=~\xa8Y\xcb\
<\x9ao\x8e\xd8A\xa2\xcf\xa0\xc7\xb4\xf6\xaa\xff;\xf7\
\xea\x14\x9b\xb6\x110m\xe2\xe2\xeefy\xd5\x8f\xf2a\
Q}\x98\xe9g\x9bu\x02\x89^] \x8c\xdbc\x7f\
\xce\xa1\x96\xfa\x00\xc0\xf4\x0a)\xc1\xda\xac\x04e\xb7D\
\x8a\xdbX\xac\x92\xc4Va\xcby5i\xd59P(\
\xcd\xd2n\xbd\x1d\x0f\xf8\x06J5Q\xb4\x0fw\x85~\
\x0dsj\x16w7%T\xfe\x1d\xd8\x98\xe4y\xbac\
Z\x848\xbd\x05\xec\xd8A@\xdd\x8a\x1b\x87n\x18\xb6\
\xcb\xec\xc9\xc1\xea\xb9\x00W\xd1\xc4=\xcc\x04l\x1c_\
\x0b\xca\x83\x9fG~Z\x0eh[Q`\x0cHP\xb8\
\xf22X544M]\xe6\xcc\x8c\xbfV,\x0b?\
\x89\xe2\xfd\xec\xbbw\xfc7\xff\xe7\x8d\xf2\x11\xa0\xfd\xee\
Zk\xe6\xacUb\xd6<\x80\x80\x09D\x94\x8cXG\
\xa7)\xd5\x22[\xa5\x85\xe2\xa1\xb7\xceWe\xb4\xb0\xc8\
\x89S\x17\xb9\xa50d\x00\xeb\xd1\xb6\xd0P\xc1\x10\x00\
p\xdaw\xfb?\x02\xbf\x8a\x06\xc5\xcc\x8c\xb8\xeb\xdd\x00\
N\x02Y\xd6\x93\x00:\xe9O\xe2\x88C\x1du\x9a\xde\
\x99Z\x9e\xe1\xf4{\xc1\xd8\xa1&\xb0b\x1e\x83\x5c\xcb\
_\xaf\x8b\x9b\xa9 @p#\x96O\xe6\xd3\x94\xea\xa6\
\xeb\xf6\xf2\xc1\xdfa\x93y\x16\x87\xcf\xe4\xd1\xca\x1b\x18\
\xeb\xbbk\x04A\xf3\xe3h\x99\xce\x8a\xd2\xcf\xcb\xebm\
\x1e\x95\x90\xb4k\x18\xff\xb38\xd7\xcay\xcd\x93\x06+\
Hm*\xa6g\x0a\x06<&\xd50\x8d\x1e\x85CG\
\x7f\xca\x808\x0a\xfe3\x84Ce\x8d\xc7\xc7\x84\xc4\xe9\
\xbe,&\x1c\x88\x89$\x1c\x87D\x0e\xc9\xcb\x81\x1b\x8f\
\x8c\x8di\xc2\x22\xce\xb6\xb3\xba\xb2\xba\x16\xdfQ\x8c\xe6\
nH\x95Cp\xfc\x0b\xc6\x0e!\x1e\xf3\x8d\xb0K\xfa\
\x99\xe7\xd9\xc0E\xaa\xfb\xea\x14kF\xaf\x13?\xbf\xe7\
\xf91\xef\xf1\x83`\x93l\x06\xfbZ5K\x9c\x07\xb1\
w\xa4\xbff4\x07c}Z{\x14f\x0f\x83\xda\x1c\
\x10\xd0\x12\xa6q\xc2O\xab\xa10\x97xZ\x00C\x13\
?_\x0b@j\xee\xa8y\x08\xe0\x80\xd6\xbeT1\xcf\
\x01h\x9c\x02\xb0\x1a\xea\xe9\x01\xfc\xba\x1e\xc8\xe8\x8e\xd1\
C\x00\x07\xb4\xf6\xb4\x94\x9e\x01\xd0\xa2'\xf0\xabF:\
\xc4\xef\x00g\xba\xa3tl\x182\x1e\xd9\x91\x19X2\
\x91\x01\xad\xf5\x9a\xd1\x10\xdf\x92\x09\xcf\x03h\xba;\xd3\
\x1d\x02H\xa91\x04\x90\xba\xae\x04\xc0A\xb7\xe6\xd8\xdc\
\x95\x008\x1e\xe2\xdb\x06\xf0\xdb\x8e`j\xd9;jI\
v\x91!q\xfcj\xb9\xdb\xa7GA<\xd8\xb0\x0f\x82\
\xb3\x1e\xec\xcf\x06\xa2m\xeel\xc9>b\xcb\xf6\x11\xdb\
<\x15Kg7\xe2j\xa8?\x1b\x80\x9e\xbd\xf3$>\
\xe8\xc9\x5c\xd0\xb3\xcf\x00x\xd2\x05\xab\x91\xfe\xa4Q\xac\
9\xcf\x1f\xc7\xd3e\xef\xc4\xe6\xdc1Q\xefT\xc6<\
z\x94\xf3Q\xbc\x03\x9e\x5c\x9d\xb2L\xd7\xdaT\x04Q\
Z\xe0oJ\xf0\xd7\x03\xf5{\x8e\xb1\x89s\xcc1\x8c\
\xe14Mg\x08\x03\x93\xa6~\x83\xfd\xafM\x0d\x0c\xd9\
\xd6F\x88y\xb8\xb3\xd5g\xe6s\x8e\x9f\xeb\x0b|\xac\
\xf1(C''K\xd1/\xf07\xe3\xcb\xf6\xbcK\xcd\
LO\x9c\xd85c\x9d1\xf5\xdd\x0d\xbe8~zP\
\x05\x90\x9b<~\xf9\xe2\xf0\xe5\xd1\xab\xe7\x8ed\xc0\xeb\
\x1f\x8ak\xaa\xa6\xad[\xca[\xc5\x82/\xdbu\x98U\
\x11\x98\xa9[\xaa\xedV4Z=z\xb6\xf2\xa0\x10j\
(\x81b\xa8\xd4\xd1MF\xa8\xa3@\xd9l\xd4\x1d=\
\xe8\x88\x97\xeanU\xf36\x9c\xaaQ5jM\x8f\xfa\
Q0\xff\xac\x80\xf4\xb7\x0a\xb5\xd8`\x5c\xcaH\xddY\
hE\x8dZ\xdf_F\x91\x80\xf62i\xfd*\xe9R\
\xcf\xb1\x8f/\x10\xfbd\x9e\xc5\xc5Y\xb7Y\x0e\xb5\xfa\
;J\xfd\x95h\xfd\xdc\x1d\x0f\xc0\xb5\xea\x97`\xa2Y\
\xe9\x97\x0a\xbd\x83\xea\xc4/\xf3h\xf7\x92\xe8\x8cRf\
P\xd5\x80O\xfb`y\xbaG\x09\xb8\x8d\xea\x10\xdd\xf5\
L\xcf\xf1^\x1dx\xad\xa2\x04\xb1_\x14\xe0\xb7\xf8\x0a\
kpj!lO\x5c\xdd\x04\xac\x99n:\xca\x0f\xe2\
I\xc5\x8bb\x88\x0f\xd5\x19QMK\xa7Lq\x98N\
\x98*\xae\xfdF\x97\xe8\x0evc\xae*\xaeu[O\
\xec\xc1I\x09*X\xbd\x81\xfb2U\xa9\x87\x1aX\x8e\
\x8e\x9aZ\xba\xe3\xa8x\xc1\xc1\xe0\x1f\xf8\xa6WO\xa3\
\xdf$\xf4l\xfbIU\xa1_\xacJ\xa2\x10\xc3\xd5\x1d\
Z\x8f\xe7\xcb\xe6o\xebD )E\xceP5fc\
\xd71\xdeZ\xd7\xcf\x94*\xcb.P\xd6s+\xd0\xfc\
\x03\xd0\xe0\xd6\xae\x9e\x7f8@\xad\xee%\xd5\xc2\xbc@\
\x0b\xc4\xcaV\x05p\x0dd\x9e\x80\xcck\xe7\x8eh\xa9\
\x88\x80\x0c3\x01\x98jH\xb0\x16\x1dMU`\xc7\xa4\
\xfaZ\x17\xea\xeb\xa8\x1e:\xf9I\xe4\xc6\xb8i\xa2\x09\
\xafr\xec\xec\x0b<\xbf\xd2\x05\xfd\xc4\xac\x82\x94\x89'\
\xd6\x8c\x89A\xa8\xd6\x9a\x1e\x22\xa7\xf5\xa1\x1bb^!\
w\x1c8\xe7b\xe0(C\xf5\x8e\x03g\xf7\xee\xc6\xc0\
\xd9R]\xdcK\x5c\xdf\x02\xc8Tq\xad|n\x14m\
\xb0\xcdPS\xa7\x1e\xa0\x06\x905\xeb\x05\xa8\xe0\xc2\x06\
\x05_\x9e\xccO\xb1\x07x\xa0nI\xd5\xf4.P\xd3\
e\xba\xab\xe2E\x02\x98\xab\x13[\x15\xb8\xb9Hn\x9d\
\xbe\xa2\xcb\x17W\xe3\x02\x1f\x83=\xd6\xf2Tq\xad}\
\xcc\x16\xd3n\xcc\xc3\x98nx\xaa\x83N\xe6P\xdd\x03\
\xbc\xf0Z7\x22Y\x15\x1c\xb2\x9e\x9dd\xb9\xba\xe4\x02\
\xc8\x98\xa5\x1b\x8e\x8a\xd7\xa7\x02\x8d^\x00\x1az\x96\x98\
\x19\xab1\x1b\x85\x17\xf1t\x1b\xa6\xce\x102\xd7U\xc5\
\xb5Y_\x99\x8epRK\xd2\xaf\x13+W\xf5\xa2]\
\x80@\xb6\xa0\x9a\xb0.\xd4a\xd9\x06\x1f\xa9\x91\xb1T\
\xe3 (+\xb2\x5c\x0b\xf3B/\xb31 \x99#E\
\xcc\xc1\xd0\x02h\x5cg\x10\x92\xe8Z\xe0T\x08\xa7-\
\xd9\x01zR\xe5\xaaZ\x17y\x18\x86$\xa8p\xca\xbd\
<U\xac\x12\x8fp\xaf\xael\xbd43\xb5\x8ef\xa6\
\xef\xa1`\x8c\xc5\xaf[\x9e\xa2|\xfdc\xca\x9e^-\
\xc9\xae\xdb_R\xe1\x8f\xa9\x9e\xa5\x98\x94\x94<NU\
\xf0\xb8\xccv\x0c\x86\x8fL\x14\x1d\xd53>T\xa5\x0e\
pa\x83\xd74\xf4\xba\xb8B\x8e\xdb/r\x88\xcd\x98\
\xa8qTJ\x9a\x1e\xf0`z\xbd\xe2\xa67\x0a\x966\
\xa2\x0b\x0c/\xea\x1a\xe7\xa0\xaa\x11/< \xbd\x97\x1d\
\x8c\x04Y\x9aBs\x96k\xc1&\x7f\xf0\xcbM\xce\xf1\
\x87v\x87\xce\xdefn\x86k\x0d\xb1H\xc4\xb8\x14\xe2\
j\xa4\xbfVO@\xdc\x83\xfe\x0f\xb0_6E\x1fr\
)\x0d\x03NN\xab\xe6\xf8\x1e\x00\xe9P\x05\xa9\x16b\
Gq\xd5\xc1\x1exO*P\xfb\x82zr*1\x9f\
\xff\xd0\xb3\x0c\xf3k\x1cf\xfc.\xe3\x9d\xe9!\xab\x82\
]C\xe2\xf2h\x8f\xbe\xcf\x1f\xb5\x8f&5\x10\xba|\
c\x22\xa9\xcb\x1f\xba\xccY\x8f\xf9\xfc,\xc7V_g\
\xa5\x19/\xd3\xade~\xcf\xf9\x04;~\xb2%\xfe\xdf\
\xcf\xb7\xb0\x03|\xe5\x08\xec\x0e\xbd\xaah`x\xc7L\
\x5c\x96\x19~\x8f\x8f\xb8\xc4\xf1\x16S-\xd8\xf1\x0d\xbc\
\xf6|\x1f\xcb* \xe2i\x16\x137\xd0\xdc.\xf4\xb5\
D\x5c\xe3\xf1\xe8\xaa>\xb8R~Q\x12\x90K\x1e\xb7\
k\x0c\x14y\xdc\xb2{<\xb0k\xcd\x06\xfb\xdc`\x07\
t\x8f\xee@\xa6!\x7f?tb\x11kr\x1d\xfc\x9d\
\xe9\xdd\xd5\xff\x00\x92*\xbe\xc7\
"

qt_resource_name = b"\
\x00\x0d\
\x01a\xd0\x87\
\x00m\
\x00a\x00i\x00n\x00_\x00i\x00c\x00o\x00n\x00.\x00s\x00v\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x8d\xcbE['\
"


def qInitResources():
    QtCore.qRegisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)


def qCleanupResources():
    QtCore.qUnregisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)


qInitResources()



import logging
import sys
import traceback
from typing import Sequence, Optional

from PySide6.QtCore import Slot, Signal, Qt, QUrl, QObject
from PySide6.QtGui import QDesktopServices, QKeySequence
from PySide6.QtWidgets import *
from qasync import asyncSlot

import qcanvas.db.database as db
from qcanvas.ui.menu_bar.grouping_preferences_menu import GroupingPreferencesMenu
from qcanvas.ui.menu_bar.theme_selection_menu import ThemeSelectionMenu
from qcanvas.ui.status_bar_reporter import StatusBarReporter
from qcanvas.ui.viewer.course_list import CourseList
from qcanvas.ui.viewer.file_list import FileRow
from qcanvas.ui.viewer.file_view_tab import FileViewTab
from qcanvas.ui.viewer.page_list_viewer import AssignmentsViewer, PagesViewer, LinkTransformer
from qcanvas.util import self_updater
from qcanvas.util.app_settings import settings
from qcanvas.util.constants import app_name
from qcanvas.util.course_indexer import DataManager
from qcanvas.util.helpers.qaction_helper import create_qaction

_aux_settings = settings.auxiliary
_no_course_selected_text = "No course selected"

class AppMainWindow(QMainWindow):
    logger = logging.getLogger()
    loaded = Signal()
    files_grouping_preference_changed = Signal(db.GroupByPreference)

    def __init__(self, data_manager: DataManager, parent: QWidget | None = None):
        super().__init__(parent)

        self.selected_course: db.Course | None = None
        self.courses: Sequence[db.Course] = []
        self.resources: dict[str, db.Resource] = {}
        self.data_manager = data_manager
        self.link_transformer = LinkTransformer(self.data_manager.link_scanners, self.resources)

        right_splitter = QSplitter()
        right_splitter.setOrientation(Qt.Orientation.Vertical)

        self.sync_button = QPushButton("Synchronize")
        self.sync_button.clicked.connect(self.sync_data)

        self.course_list = CourseList(self.data_manager)
        self.course_list.course_selected.connect(self.on_course_selected)

        self.assignment_viewer = AssignmentsViewer(self.link_transformer)
        self.assignment_viewer.viewer.anchorClicked.connect(self.viewer_link_clicked)

        self.pages_viewer = PagesViewer(self.link_transformer)
        self.pages_viewer.viewer.anchorClicked.connect(self.viewer_link_clicked)

        self.file_viewer = FileViewTab(data_manager.download_pool)
        self.file_viewer.files_column.tree.itemActivated.connect(self.download_file_from_file_pane)
        self.file_viewer.assignment_files_column.tree.itemActivated.connect(self.download_file_from_file_pane)

        self.tab_widget = QTabWidget()
        self.tab_widget.insertTab(0, self.file_viewer, "Files")
        self.tab_widget.insertTab(1, self.assignment_viewer, "Assignments")
        self.tab_widget.insertTab(2, self.pages_viewer, "Pages")

        self.course_name_label = QLabel(_no_course_selected_text)
        self.course_name_label.setStyleSheet("font-weight: bold;")
        course_stack_layout = self.create_layout_and_add_widgets(QVBoxLayout, self.course_name_label, self.tab_widget)

        h_layout = self.create_layout_and_add_widgets(QHBoxLayout, self.course_list, course_stack_layout)
        h_layout.setStretch(1, 1)

        widget = QWidget()
        widget.setLayout(self.create_layout_and_add_widgets(QVBoxLayout, h_layout, self.sync_button))
        self.setCentralWidget(widget)

        self.setup_menu_bar()

        self.files_grouping_preference_changed.connect(self.on_grouping_preference_changed)

        self.loaded.connect(self.load_course_list)
        self.loaded.connect(self.check_for_update)
        self.loaded.emit()

        self.restore_window_position()

        # Activate the statusbar so it doesn't just appear randomly later
        bar: QStatusBar = self.statusBar()
        # Set its height so it doesn't get bigger when there's a progress bar in it
        bar.setFixedHeight(bar.height())

    @staticmethod
    def create_layout_and_add_widgets(layout_type: type, *widgets) -> QLayout:
        layout = layout_type()

        for widget in widgets:
            if isinstance(widget, QLayout):
                layout.addLayout(widget)
            else:
                layout.addWidget(widget)

        return layout

    def setup_menu_bar(self):
        menu_bar = self.menuBar()

        app_menu: QMenu = menu_bar.addMenu("App")
        view_menu: QMenu = menu_bar.addMenu("View")

        app_menu.addAction(self.setup_quick_authentication_action(app_menu))
        app_menu.addMenu(ThemeSelectionMenu())
        view_menu.addMenu(self.setup_group_by_menu())

    def setup_group_by_menu(self) -> QMenu:
        file_grouping_menu = GroupingPreferencesMenu(self.data_manager)
        self.course_list.course_selected.connect(file_grouping_menu.course_changed)
        file_grouping_menu.preference_changed.connect(self.on_grouping_preference_changed)

        return file_grouping_menu

    def setup_quick_authentication_action(self, parent: QObject):
        return create_qaction(
            name="Quick canvas login",
            shortcut=QKeySequence("Ctrl+O"),
            triggered=self.open_quick_auth_in_browser,
            parent=parent
        )

    @asyncSlot()
    async def open_quick_auth_in_browser(self):
        opening_progress_dialog = QProgressDialog("Opening canvas", None, 0, 0, self)
        opening_progress_dialog.setWindowTitle("Please wait")
        opening_progress_dialog.show()
        QDesktopServices.openUrl(await self.data_manager.client.get_temp_session_link())
        opening_progress_dialog.close()

    def closeEvent(self, event):
        settings.geometry = self.saveGeometry()
        settings.window_state = self.saveState()

    def restore_window_position(self):
        self.restoreGeometry(settings.geometry)
        self.restoreState(settings.window_state)

    @asyncSlot(QUrl)
    async def viewer_link_clicked(self, url: QUrl):
        # The url of a transformed link will start with a specific prefix
        if url.toString().startswith(LinkTransformer.transformed_url_prefix):
            # The rest of the 'url' is just the file id
            resource = self.resources[url.toString().removeprefix(LinkTransformer.transformed_url_prefix)]

            await self.data_manager.download_resource(resource)
            QDesktopServices.openUrl(QUrl.fromLocalFile(resource.download_location.absolute()))
        else:
            QDesktopServices.openUrl(url)

    @asyncSlot(QTreeWidgetItem, int)
    async def download_file_from_file_pane(self, item: QTreeWidgetItem, _: int):
        if isinstance(item, FileRow):
            await self.data_manager.download_resource(item.resource)
            QDesktopServices.openUrl(QUrl.fromLocalFile(item.resource.download_location.absolute()))

    @asyncSlot()
    async def sync_data(self):
        self.sync_button.setEnabled(False)
        self.sync_button.setText("Synchronizing")
        try:
            await self.data_manager.synchronize_with_canvas(StatusBarReporter(self.statusBar()))
            await self.load_course_list()

        finally:
            self.sync_button.setEnabled(True)
            self.sync_button.setText("Synchronize")

    @asyncSlot()
    async def load_course_list(self):
        self.courses = (await self.data_manager.get_data())
        self.selected_course = None
        self.resources.clear()

        for course in self.courses:
            self.resources.update({resource.id: resource for resource in course.resources})

        self.course_list.load_course_list(self.courses)

    @asyncSlot()
    async def check_for_update(self):
        try:
            newer_version, installed_version = await self_updater.get_newer_version()

            if newer_version is not None and newer_version != settings.ignored_update:
                msg_box = QMessageBox(
                    QMessageBox.Icon.Question,
                    "Update available",
                    f"There is an update available ({installed_version} -> {newer_version})\nDo you want to update?\nThe program will close after the update is finished.",
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                    self
                )

                def ignore_update():
                    settings.ignored_update = newer_version

                msg_box.accepted.connect(self.do_self_update)
                msg_box.rejected.connect(ignore_update)
                msg_box.show()
            else:
                print("No update available (or skipping this update)")
        except BaseException as e:
            sys.stderr.write(f"Could not check for updates: {e}\n")
            traceback.print_exc()
            sys.stderr.write("This can be ignored if in a dev environment\n")

    @asyncSlot()
    async def do_self_update(self):
        try:
            progress_diag = QProgressDialog("Updating", None, 0, 0, self)
            progress_diag.setWindowTitle(app_name)
            progress_diag.show()
            await self_updater.do_update()
            self.close()
        except BaseException as e:
            traceback.print_exc()

            QMessageBox(
                QMessageBox.Icon.Critical,
                "Error",
                "An error occurred during the update",
                parent=self
            ).show()

    @Slot(db.Course)
    def on_course_selected(self, course: Optional[db.Course]):
        if course is not None:
            self.selected_course = course
            # todo these should really be slots connected to this signal...
            self.pages_viewer.fill_tree(course)
            self.assignment_viewer.fill_tree(course)
            self.file_viewer.load_course_files(course)
            self.course_name_label.setText(course.name)
        else:
            self.selected_course = None
            self.file_viewer.clear()
            self.pages_viewer.clear()
            self.assignment_viewer.clear()
            self.course_name_label.setText(_no_course_selected_text)

    @asyncSlot(db.CoursePreferences)
    async def on_grouping_preference_changed(self):
        self.file_viewer.load_course_files(self.selected_course)



from abc import ABC, abstractmethod
from typing import AsyncIterator

import httpx
from bs4 import Tag

import qcanvas.db as db


class ResourceScanner(ABC):
    """
    A resource scanner extracts resources from canvas pages.
    The resource may be an embedded video, a file or anything that will be of use to the user.
    Each scanner should be responsible for only 1 type of resource.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of the resource scanner.
        Will be attached to the resource id externally.
        """
        ...

    @abstractmethod
    def accepts_link(self, link: Tag) -> bool:
        """
        Whether this resource scanner accepts the specified link
        """
        ...

    @abstractmethod
    async def extract_resource(self, link: Tag, file_id: str) -> db.Resource:
        """
        Extract information about the resource in the specified tag
        Parameters
        ----------
        link
            The element that links to the resource
        file_id
            The id of the file (as produced from `extract_id`)
        Returns
        -------
            The resource
        """
        ...

    @abstractmethod
    def extract_id(self, link: Tag) -> str:
        """
        Extracts a unique id from a file link
        """
        ...

    async def download(self, resource: db.Resource) -> AsyncIterator[int]:
        yield 0

        download_destination = resource.download_location
        download_destination.parent.mkdir(parents=True, exist_ok=True)

        with open(download_destination, "wb") as file:
            async with httpx.AsyncClient(follow_redirects=True) as client:
                async with client.stream(method='get', url=resource.url) as resp:
                    resp.raise_for_status()

                    async for chunk in resp.aiter_bytes():
                        file.write(chunk)
                        yield resp.num_bytes_downloaded



from .database import Resource, Module, ModuleItem, ModulePage, ModuleFile, ResourceState, Course, Term, \
    Assignment, Base, PageLike, ResourceToModuleItemAssociation, ResourceToAssignmentAssociation, CoursePreferences, \
    GroupByPreference
from .db_converter_helper import convert_course, convert_page, convert_file, convert_legacy_file, \
    convert_assignment, convert_module, convert_term, convert_file_page



import json
from typing import Any

from bs4 import Tag, BeautifulSoup
from httpx import AsyncClient

from qcanvas import db as db
from qcanvas.util.link_scanner import ResourceScanner


class CanvasMediaObjectScanner(ResourceScanner):

    def __init__(self, client: AsyncClient):
        self.client = client

    @property
    def name(self) -> str:
        return "canvas_media_object"

    def accepts_link(self, link: Tag) -> bool:
        return (
                link.name == "iframe"
                and "data-media-type" in link.attrs.keys()
                and link.attrs["data-media-type"] == "video"
        )

    async def extract_resource(self, link: Tag, file_id: str) -> db.Resource:
        # Get the page for the embedded player (I could not find another way to get the needed data from canvas)
        response = (await self.client.get(link.attrs["src"])).text
        # Parse the HTML response
        doc = BeautifulSoup(response, "html.parser")
        media_data: None | dict[str, Any] = None

        # Find all script tags (one of them has the data we are interested in)
        for script_tag in doc.find_all("script", {}):
            body = script_tag.text.strip()

            # If the tag content starts with this then it has the data we want
            if "INST = {" in body:
                # Find the data that we are interested in (is on a line that starts with "ENV = ")
                line: str = next(filter(lambda x: x.strip().startswith("ENV ="), script_tag.text.splitlines()))
                # Parse the json embedded in the script tag
                media_data = json.loads(line.lstrip("ENV = ").rstrip(";"))["media_object"]
                break

        if media_data is None:
            raise Exception("Could not extract media info")

        # The highest quality stream is the first
        media_source = media_data["media_sources"][0]

        return db.Resource(
            id=file_id,
            url=media_source["src"],
            file_name=media_data["title"],
            file_size=int(media_source["size"]) * 1024  # Seems to be recorded in KiB, not bytes
        )

    def extract_id(self, link: Tag) -> str:
        return link.attrs["data-media-id"]



from datetime import datetime
from typing import Sequence

from PySide6.QtCore import QItemSelection, Slot, Signal, QObject
from PySide6.QtGui import QStandardItemModel, QStandardItem
from PySide6.QtWidgets import QTreeView
from qasync import asyncSlot

import qcanvas.db as db
from qcanvas.util.course_indexer import DataManager


class CourseNode(QStandardItem, QObject):
    name_changed = Signal(db.Course, str)

    def __init__(self, course: db.Course):
        QObject.__init__(self)
        QStandardItem.__init__(self, course.preferences.local_name or course.name)
        self.course = course

    def setData(self, value, role=...):
        if isinstance(value, str):
            value = value.strip()

            if len(value) == 0:
                super().setData(self.course.name, role)
                self.name_changed.emit(self.course, None)
            else:
                super().setData(value, role)
                self.name_changed.emit(self.course, value)
        else:
            super().setData(value, role)


class CourseList(QTreeView):
    course_selected = Signal(db.Course)

    def __init__(self, data_manager: DataManager):
        super().__init__()
        self.data_manager = data_manager
        self.model = QStandardItemModel()
        self.setModel(self.model)
        self.selectionModel().selectionChanged.connect(self._on_selection_changed)

    @asyncSlot(db.Course, str)
    async def course_name_changed(self, course: db.Course, name: str):
        course.preferences.local_name = name
        await self.data_manager.update_item(course.preferences)

    def load_course_list(self, courses: Sequence[db.Course]):
        self.model.clear()
        self.model.setHorizontalHeaderLabels(["Course"])

        courses_root = self.model.invisibleRootItem()

        for term, courses in self.group_courses_by_term(courses):
            term_node = QStandardItem(term.name)
            term_node.setEditable(False)
            term_node.setSelectable(False)

            for course in courses:
                course_node = CourseNode(course)
                course_node.name_changed.connect(self.course_name_changed)
                term_node.appendRow(course_node)

            courses_root.appendRow(term_node)

        self.expandAll()

    @staticmethod
    def group_courses_by_term(courses: Sequence[db.Course]):
        courses_grouped_by_term: dict[db.Term, list[db.Course]] = {}

        # Put courses into groups in the above dict
        for course in courses:
            if course.term in courses_grouped_by_term:
                courses_grouped_by_term[course.term].append(course)
            else:
                courses_grouped_by_term[course.term] = [course]

        # Convert the dict item list into a mutable list
        pairs = list(courses_grouped_by_term.items())
        # Sort them by start date, with most recent terms at the start
        pairs.sort(key=lambda x: x[0].start_at or datetime.min, reverse=True)

        return pairs

    @Slot(QItemSelection, QItemSelection)
    def _on_selection_changed(self, selected: QItemSelection, deselected: QItemSelection):
        if len(self.selectedIndexes()) == 0:
            self.course_selected.emit(None)
        else:
            item = self.model.itemFromIndex(self.selectedIndexes()[0])

            if isinstance(item, CourseNode):
                self.course_selected.emit(item.course)



from PySide6.QtCore import Slot, Signal
from PySide6.QtWidgets import QMenu, QWidget
from qasync import asyncSlot

import qcanvas.db as db
from qcanvas.util.course_indexer import DataManager
from qcanvas.util.helpers.qaction_helper import create_qaction


class GroupingPreferencesMenu(QMenu):
    _preference_changed_private = Signal(db.GroupByPreference)
    preference_changed = Signal()

    def __init__(self, data_manager: DataManager, parent: QWidget | None = None):
        super().__init__("Group files by", parent)
        self.setEnabled(False)

        self.data_manager = data_manager
        self._selected_course: db.Course | None = None

        self.group_by_modules_action = self._make_action("Modules", db.GroupByPreference.GROUP_BY_MODULES)
        self.group_by_pages_action = self._make_action("Pages", db.GroupByPreference.GROUP_BY_PAGES)

        self.addActions([self.group_by_pages_action, self.group_by_modules_action])

        self._preference_changed_private.connect(self._on_preference_changed)

    def _make_action(self, text: str, preference_value: db.GroupByPreference):
        return create_qaction(
            name=text,
            parent=self,
            triggered=lambda: self._preference_changed_private.emit(preference_value),
            checkable=True,
            checked=False
        )

    @Slot()
    def course_changed(self, course: db.Course | None):
        if course is not None:
            self._selected_course = course
            self._update_actions()
            self.setEnabled(True)
        else:
            self._selected_course = None
            self.setEnabled(False)

    @asyncSlot()
    async def _on_preference_changed(self, preference: db.GroupByPreference):
        if self._selected_course is not None:
            self._selected_course.preferences.files_group_by_preference = preference
            self._update_actions()
            # todo not sure if this should go in main_ui or here...
            await self.data_manager.update_item(self._selected_course.preferences)

            self.preference_changed.emit()

    def _update_actions(self):
        preference = self._selected_course.preferences.files_group_by_preference

        self.group_by_pages_action.setChecked(preference == db.GroupByPreference.GROUP_BY_PAGES)
        self.group_by_modules_action.setChecked(preference == db.GroupByPreference.GROUP_BY_MODULES)



default_assignments_module_names = ["assessments", "assessment"]
app_name = "QCanvas"
package_name = "qcanvas"
# Passed back to the launcher script to indicate that the program should be restarted (after an update)
updated_and_needs_restart_return_code = 144



import asyncio
import logging
import traceback
from asyncio import Task
from dataclasses import dataclass
from typing import Sequence

from gql import gql
from sqlalchemy import select
from sqlalchemy.dialects.sqlite import insert as sqlite_upsert
from sqlalchemy.ext.asyncio.session import async_sessionmaker as AsyncSessionMaker, AsyncSession
from sqlalchemy.orm import selectin_polymorphic, selectinload

import qcanvas.db as db
import qcanvas.queries as queries
import qcanvas.util.course_indexer.conversion_helpers as conv_helper
import qcanvas.util.course_indexer.resource_helpers as resource_helper
from qcanvas.net.canvas import CanvasClient
from qcanvas.util.download_pool import DownloadPool
from qcanvas.util.link_scanner.canvas_link_scanner import canvas_resource_id_prefix
from qcanvas.util.link_scanner.resource_scanner import ResourceScanner
from qcanvas.util.progress_reporter import ProgressReporter, noop_reporter
from qcanvas.util.task_pool import TaskPool

_logger = logging.getLogger("course_loader")


@dataclass
class TransientModulePage:
    page: queries.Page | queries.File
    course_id: str
    module_id: str
    position: int


def _prepare_out_of_date_pages_for_loading(g_courses: Sequence[queries.Course], pages: Sequence[db.ModuleItem]) -> list[
    TransientModulePage]:
    """
    Removes pages that are up-to-date from the pages list by comparing the last update time of the pages from the query
    to the last update time of the pages in the database.

    Parameters
    ----------
    g_courses
        The list of courses (with module items) to check for an update.
    pages
        The list of pages already existing in the database.
    Returns
    -------
    list[TransientModulePage]
        A list of pages that have had an update to them.
    """
    pages_id_mapped = {x.id: x for x in pages}

    result: list[TransientModulePage] = []

    for g_course in g_courses:
        for g_module in g_course.modules_connection.nodes:
            for item_position, g_module_item in enumerate(g_module.module_items):
                content = g_module_item.content

                if isinstance(content, (queries.File, queries.Page)):
                    # todo need to decide how to only rescan old pages or only rescan new pages without fetching content of old pages again for no good reason
                    if (
                            content.m_id not in pages_id_mapped
                            or content.updated_at.replace(tzinfo=None) > pages_id_mapped[content.m_id].updated_at
                    ):
                        result.append(TransientModulePage(content, g_course.m_id, g_module.q_id, item_position))
                    else:
                        _logger.debug("Page %s is already up to date", content.m_id)

    return result


# todo make this reusable and add some way of refreshing only a list of pages or one page or one course or something
# todo use logger instead of print and put some signals around the place for useful things, e.g. indexing progress
class DataManager:
    """
    Responsible for storing all data pulled from canvas or other websites in the database.
    Provides functions for synchronizing with canvas and downloading files.
    """
    def __init__(self,
                 client: CanvasClient,
                 sessionmaker: AsyncSessionMaker,
                 link_scanners: Sequence[ResourceScanner]):

        self.client = client
        self._link_scanners = link_scanners
        self._session_maker = sessionmaker

        self._resource_pool = TaskPool[db.Resource]()
        self._moduleitem_pool = TaskPool[db.ModuleItem]()
        self.download_pool = DownloadPool()

        # Map all the scanners we have to their own name
        self._scanner_name_map = {scanner.name: scanner for scanner in self._link_scanners}

        self._init_called = False

    async def init(self):
        """
        Load existing pages and resources from the database, so they don't have to be fetched from canvas again
        """
        self._init_called = True

        async with self._session_maker.begin() as session:
            # Load existing pages and resources from the database
            existing_pages = (await session.execute(
                select(db.ModuleItem)
                .options(selectin_polymorphic(db.ModuleItem, [db.ModulePage]))
            )).scalars().all()

            existing_resources = (await session.execute(
                select(db.Resource)
            )).scalars().all()

            # Add the existing items to the relevant taskpools
            self._add_resources_and_pages_to_taskpool(existing_pages=existing_pages,
                                                      existing_resources=existing_resources)

    async def _download_resource_helper(self, link_handler: ResourceScanner, resource: db.Resource):
        try:
            async for progress in link_handler.download(resource):
                yield progress

            # Do this here because this function will only be called once for this resource
            async with self._session_maker.begin() as session:
                session.add(resource)
                resource.state = db.ResourceState.DOWNLOADED
        except BaseException as e:
            # Something went wrong, record the failure in the database
            async with self._session_maker.begin() as session:
                session.add(resource)
                resource.state = db.ResourceState.FAILED
                resource.fail_message = str(e)

            raise e

    async def download_resource(self, resource: db.Resource):
        if not self._init_called:
            raise Exception("Init was not called")

        # Resource ids look like this: "canvas_file:387837", and we just want the "canvas_file" part
        scanner_name: str = resource.id.split(':', 2)[0]
        # Find the scanner that will deal with this resource
        scanner = self._scanner_name_map[scanner_name]

        await self.download_pool.submit(resource.id, lambda: self._download_resource_helper(scanner, resource))

    async def update_item(self, item: db.Base):
        async with self._session_maker.begin() as session:
            await session.merge(item)

    async def get_data(self):
        """
        Loads all the course data
        """
        async with self._session_maker.begin() as session:
            module_items_load = selectinload(db.Course.modules).joinedload(db.Module.items)

            # Eagerly load fucking everything
            options = [
                selectinload(db.Course.modules)
                .joinedload(db.Module.course),

                module_items_load.selectin_polymorphic([db.ModulePage, db.ModuleFile])
                .joinedload(db.ModuleItem.module),

                module_items_load.joinedload(db.ModuleItem.resources),
                selectinload(db.Course.assignments)
                .joinedload(db.Assignment.course),

                selectinload(db.Course.assignments)
                .joinedload(db.Assignment.resources),

                selectinload(db.Course.term),

                selectinload(db.Course.module_items)
                .joinedload(db.ModuleItem.course),

                selectinload(db.Course.preferences)
                .joinedload(db.CoursePreferences.course),

                selectinload(db.Course.resources)
                .joinedload(db.Resource.course)
            ]

            return (await session.execute(select(db.Course).options(*options))).scalars().all()

    async def synchronize_with_canvas(self, progress_reporter: ProgressReporter = noop_reporter):
        section = progress_reporter.section("Loading index", 0)
        raw_query = (await self.client.do_graphql_query(gql(queries.all_courses.DEFINITION), detailed=True))
        section.increment_progress()

        await self.load_courses_data(queries.AllCoursesQueryData(**raw_query).all_courses, progress_reporter)

    async def load_courses_data(self, g_courses: Sequence[queries.Course], progress_reporter: ProgressReporter):
        """
        Loads data for all specified courses, including loading module pages and scanning for resources.
        """

        if not self._init_called:
            raise Exception("Init was not called")

        try:
            async with self._session_maker.begin() as session:
                # Load module pages/files for the courses
                await self._load_module_items(g_courses, session, progress_reporter)

                # Collect assignments from the courses
                assignments = []

                for g_course in g_courses:
                    # Create needed data in the session
                    term = await conv_helper.create_term(g_course, session)
                    await conv_helper.create_course(g_course, session, term)
                    await conv_helper.create_modules(g_course, session)

                    # Add course assignments to the list
                    assignments.extend(await conv_helper.create_assignments(g_course, session))

                # Scan assignments for resources
                await self._scan_assignments_for_resources(assignments, session, progress_reporter)

                # Add all resources back into the session
                session.add_all(self._resource_pool.results())
                progress_reporter.finished()
        except BaseException as e:
            traceback.print_exc()
            progress_reporter.errored(e)

    async def _scan_assignments_for_resources(self, assignments: Sequence[db.Assignment], session: AsyncSession,
                                              progress_reporter: ProgressReporter):
        """
        Scans assignments for resources
        """

        # Link the resources found to each page in the database
        await resource_helper.create_assignment_resource_relations(
            # Find all the resources in each assignment description
            await resource_helper.find_resources_in_pages(
                link_scanners=self._link_scanners,
                resource_pool=self._resource_pool,
                items=assignments,
                progress_reporter=progress_reporter
            ),
            session
        )

    async def _load_module_items(self, g_courses: Sequence[queries.Course], session: AsyncSession,
                                 progress_reporter: ProgressReporter):
        # Get the ids of all the courses we are going to index/load
        course_ids = [g_course.m_id for g_course in g_courses]

        # Prepare pages for loading
        existing_pages = (
            await session.execute(
                select(db.ModulePage)
                .where(db.ModuleItem.course_id.in_(course_ids))
            )).scalars().all()

        # Filter out pages that don't need updating
        pages_to_update = _prepare_out_of_date_pages_for_loading(g_courses, existing_pages)

        if len(pages_to_update) == 0:
            return

        # Load the content for all the pages that need updating
        module_items: list[db.ModuleItem] = await self._load_content_for_pages(pages_to_update, progress_reporter)
        module_pages = [item for item in module_items if isinstance(item, db.ModulePage)]

        # Link the resources found to the pages they were found on and add them to the database
        await resource_helper.create_module_item_resource_relations(
            # Find all the resources in each page
            await resource_helper.find_resources_in_pages(
                link_scanners=self._link_scanners,
                resource_pool=self._resource_pool,
                progress_reporter=progress_reporter,
                # Collect just the module pages for scanning
                items=module_pages
            ),
            session
        )

        # empty inserts/upserts causes an sql error. don't do them
        if len(module_pages) > 0:
            # Add all the module items to the session
            # shitty bandaid fix
            upsert_item = sqlite_upsert(db.ModuleItem).values([self.moduleitem_dict(item) for item in module_pages])
            upsert_item = upsert_item.on_conflict_do_update(
                index_elements=[db.ModuleItem.id],
                set_=dict(name=upsert_item.excluded.name, updated_at=upsert_item.excluded.updated_at,
                          position=upsert_item.excluded.position),

            )

            upsert_page = sqlite_upsert(db.ModulePage).values([self.page_dict(item) for item in module_pages])
            upsert_page = upsert_page.on_conflict_do_update(
                index_elements=[db.ModulePage.id],
                set_=dict(content=upsert_page.excluded.content)
            )

            await session.execute(upsert_item)
            await session.execute(upsert_page)

        session.add_all([item for item in module_items if isinstance(item, db.ModuleFile)])

    @staticmethod
    def page_dict(page: db.ModulePage) -> dict[str, object]:
        return {"id": page.id, "content": page.content}

    @staticmethod
    def moduleitem_dict(page: db.ModuleItem) -> dict[str, object]:
        return {"id": page.id, "name": page.name, "updated_at": page.updated_at, "position": page.position,
                "module_id": page.module_id, "course_id": page.course_id, "type": page.type,
                "created_at": page.created_at}

    def _add_resources_and_pages_to_taskpool(self, existing_pages: Sequence[db.ModuleItem],
                                             existing_resources: Sequence[db.Resource]):
        self._moduleitem_pool.add_values({page.id: page for page in existing_pages})
        self._resource_pool.add_values({resource.id: resource for resource in existing_resources})
        # Add downloaded resources to the resource pool so we don't download them again
        self.download_pool.add_values(
            {resource.id: None for resource in existing_resources if resource.state == db.ResourceState.DOWNLOADED})

    async def _load_content_for_pages(self, pages: Sequence[TransientModulePage],
                                      progress_reporter: ProgressReporter) -> list[db.ModuleItem]:
        """
        Loads the page content for the specified pages
        Parameters
        ----------
        pages
            The pages to load
        Returns
        -------
        list
            The list of complete pages with page content loaded.
        """
        progress = progress_reporter.section("Loading page content", len(pages))
        tasks: list[Task[db.ModuleItem | None]] = []

        for page in pages:
            content = page.page

            # Load the content for the pages
            if isinstance(content, queries.File):
                task = asyncio.create_task(
                    self._load_module_file(content, page.course_id, page.module_id, page.position))
                task.add_done_callback(progress.increment_progress)
                tasks.append(task)
            elif isinstance(content, queries.Page):
                task = asyncio.create_task(
                    self.load_module_page(content, page.course_id, page.module_id, page.position))
                task.add_done_callback(progress.increment_progress)
                tasks.append(task)

        if len(tasks) > 0:
            await asyncio.wait(tasks)

            # Collect results and filter out nulls
            return [task.result() for task in tasks if task.result() is not None]
        else:
            return []

    async def _load_module_file(self, g_file: queries.File, course_id: str, module_id: str,
                                position: int) -> db.ModuleFile:
        """
        Fetches resource information for the module file and converts it into a module item
        """
        _logger.debug(f"Loading module file %s %s", g_file.m_id, g_file.display_name)

        resource = await self._resource_pool.submit(
            f"{canvas_resource_id_prefix}:{g_file.m_id}",  # to match the format used by canvas link extractor
            lambda: self._fetch_module_file_resource(g_file, course_id)
        )

        return await self._moduleitem_pool.submit(
            g_file.m_id,
            lambda: self._fetch_module_file_page(g_file, resource, course_id, module_id, position)
        )

    async def _fetch_module_file_resource(self, file: queries.File, course_id: str) -> db.Resource:
        """
        Fetches information about the specified file from canvas
        """
        _logger.debug(f"Fetching file (for module file) %s %s", file.m_id, file.display_name)
        result = await self.client.get_file(file.m_id, course_id)
        resource = db.convert_file(file, result.size)
        resource.id = f"{canvas_resource_id_prefix}:{resource.id}"
        resource.course_id = course_id

        return resource

    async def load_module_page(self, g_page: queries.Page, course_id: str, module_id: str,
                               position: int) -> db.ModulePage | None:
        """
        Creates task for loading the specified module page
        """
        return await self._moduleitem_pool.submit(
            g_page.m_id,
            lambda: self._fetch_module_item_page(g_page, course_id, module_id, position)
        )

    async def _fetch_module_item_page(self, page: queries.Page, course_id: str, module_id: str,
                                      position: int) -> db.ModulePage | None:
        """
        Fetches module page content from canvas. Returns None if the page could not be loaded.
        """
        _logger.debug("Fetching module page %s %s", page.m_id, page.title)

        try:
            # Get the page
            result = await self.client.get_page(page.m_id, course_id)
        except BaseException as e:
            # Handle any errors
            _logger.error(e)
            traceback.print_exc()
            return None

        if result.locked_for_user:
            _logger.error("Page %s %s is locked", page.m_id, page.title)
            return None

        page = db.convert_page(page, result.body)
        page.module_id = module_id
        page.course_id = course_id
        page.position = position

        return page

    @staticmethod
    async def _fetch_module_file_page(file: queries.File, resource: db.Resource, course_id: str,
                                      module_id: str, position: int) -> db.ModuleFile:
        """
        Converts module file information into the database format
        """
        _logger.debug(f"Creating page for module file %s %s", file.m_id, file.display_name)

        page = db.convert_file_page(file)
        page.module_id = module_id
        page.course_id = course_id
        page.position = position
        page.resources.append(resource)

        return page

    @property
    def link_scanners(self):
        return self._link_scanners



"""
Generated by qenerate plugin=pydantic_v1. DO NOT MODIFY MANUALLY!
"""
from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611
from datetime import datetime  # noqa: F401 # pylint: disable=W0611
from enum import Enum  # noqa: F401 # pylint: disable=W0611
from typing import (  # noqa: F401 # pylint: disable=W0611
    Any,
    Optional,
    Union,
)

from pydantic import (  # noqa: F401 # pylint: disable=W0611
    BaseModel,
    Extra,
    Field,
    Json,
)

from qcanvas.queries.canvas_course_data import CanvasCourseData

DEFINITION = """
fragment CanvasCourseData on Course {
    _id
#    id
    name
    courseNickname
    term {
        endAt
        startAt
        name
        id
    }
    assignmentsConnection @include(if: $detailed) {
        nodes {
            description
            courseId
            dueAt
            createdAt
            id
            name
            position
            updatedAt
        }
    }
    modulesConnection @include(if: $detailed) {
        nodes {
            name
            id
            moduleItems {
                content {
                    ... on File {
                        _id
                        displayName
                        createdAt
                        updatedAt
                        url
                        size
                        mimeClass
                        contentType
                    }
                    ... on Page {
                        _id
                        title
                        updatedAt
                        createdAt
                    }
                }
            }
        }
    }
}

query AllCourses($detailed: Boolean!) {
  allCourses {
    ...CanvasCourseData
  }
}
"""


class ConfiguredBaseModel(BaseModel):
    class Config:
        smart_union = True
        extra = Extra.forbid


class AllCoursesQueryData(ConfiguredBaseModel):
    all_courses: Optional[list[CanvasCourseData]] = Field(..., alias="allCourses")


def query(query_func: Callable, **kwargs: Any) -> AllCoursesQueryData:
    """
    This is a convenience function which queries and parses the data into
    concrete types. It should be compatible with most GQL clients.
    You do not have to use it to consume the generated data classes.
    Alternatively, you can also mime and alternate the behavior
    of this function in the caller.

    Parameters:
        query_func (Callable): Function which queries your GQL Server
        kwargs: optional arguments that will be passed to the query function

    Returns:
        AllCoursesQueryData: queried data parsed into generated classes
    """
    raw_data: dict[Any, Any] = query_func(DEFINITION, **kwargs)
    return AllCoursesQueryData(**raw_data)



from datetime import datetime

import qcanvas.db.database as db
import qcanvas.queries as gql
from qcanvas.net.canvas.legacy_canvas_types import LegacyFile
from qcanvas.util.helpers.canvas_sanitiser import remove_garbage_from_title as clean_string


def convert_term(term: gql.Term) -> db.Term:
    return db.Term(
        id=term.q_id,
        name=clean_string(term.name),
        start_at=term.start_at,
        end_at=term.end_at
    )


def convert_course(course: gql.Course) -> db.Course:
    return db.Course(
        id=course.m_id,
        name=clean_string(course.name),
        local_name=clean_string(course.course_nickname)
    )


def convert_assignment(assignment: gql.Assignment) -> db.Assignment:
    return db.Assignment(
        id=assignment.q_id,
        name=clean_string(assignment.name),
        description=assignment.description,
        due_at=assignment.due_at,
        created_at=assignment.created_at,
        updated_at=assignment.updated_at,
        position=assignment.position
    )


def convert_page(page: gql.Page, content: str) -> db.ModulePage:
    return db.ModulePage(
        id=page.m_id,
        content=content,
        created_at=page.created_at,
        updated_at=page.updated_at,
        name=clean_string(page.title)
    )


def convert_file_page(file: gql.File) -> db.ModuleFile:
    return db.ModuleFile(
        id=file.m_id,
        created_at=file.created_at,
        updated_at=file.updated_at,
        name=clean_string(file.display_name)
    )


def convert_file(file: gql.File, file_size: int) -> db.Resource:
    return db.Resource(
        id=str(file.m_id),
        url=file.url,
        file_name=clean_string(file.display_name),
        date_discovered=datetime.now(),
        file_size=file_size
    )


def convert_legacy_file(file: LegacyFile) -> db.Resource:
    return db.Resource(
        id=str(file.id),
        url=file.url,
        file_name=clean_string(file.display_name),
        date_discovered=datetime.now(),
        file_size=file.size
    )


def convert_module(module: gql.Module) -> db.Module:
    return db.Module(
        id=module.q_id,
        name=clean_string(module.name)
    )



"""
Generated by qenerate plugin=pydantic_v1. DO NOT MODIFY MANUALLY!
"""
from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611
from datetime import datetime  # noqa: F401 # pylint: disable=W0611
from enum import Enum  # noqa: F401 # pylint: disable=W0611
from typing import (  # noqa: F401 # pylint: disable=W0611
    Any,
    Optional,
    Union,
)

from pydantic import (  # noqa: F401 # pylint: disable=W0611
    BaseModel,
    Extra,
    Field,
    Json,
)

DEFINITION = """
fragment CanvasCourseData on Course {
    _id
    name
    courseNickname
    term {
        endAt
        startAt
        name
        id
    }
    assignmentsConnection @include(if: $detailed) {
        nodes {
            description
            courseId
            dueAt
            createdAt
            id
            name
            position
            updatedAt
        }
    }
    modulesConnection @include(if: $detailed) {
        nodes {
            name
            id
            moduleItems {
                content {
                    ... on File {
                        _id
                        displayName
                        createdAt
                        updatedAt
                        url
                        size
                        mimeClass
                        contentType
                    }
                    ... on Page {
                        _id
                        title
                        updatedAt
                        createdAt
                    }
                }
            }
        }
    }
}
"""


class ConfiguredBaseModel(BaseModel):
    class Config:
        smart_union = True
        extra = Extra.forbid


class Term(ConfiguredBaseModel):
    end_at: Optional[datetime] = Field(..., alias="endAt")
    start_at: Optional[datetime] = Field(..., alias="startAt")
    name: Optional[str] = Field(..., alias="name")
    q_id: str = Field(..., alias="id")


class Assignment(ConfiguredBaseModel):
    description: Optional[str] = Field(..., alias="description")
    course_id: Optional[str] = Field(..., alias="courseId")
    due_at: Optional[datetime] = Field(..., alias="dueAt")
    created_at: Optional[datetime] = Field(..., alias="createdAt")
    q_id: str = Field(..., alias="id")
    name: Optional[str] = Field(..., alias="name")
    position: Optional[int] = Field(..., alias="position")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")


class AssignmentConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[Assignment]]] = Field(..., alias="nodes")


class ModuleItemInterface(ConfiguredBaseModel):
    ...


class File(ModuleItemInterface):
    m_id: str = Field(..., alias="_id")
    display_name: Optional[str] = Field(..., alias="displayName")
    created_at: Optional[datetime] = Field(..., alias="createdAt")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    url: Optional[str] = Field(..., alias="url")
    size: Optional[str] = Field(..., alias="size")
    mime_class: Optional[str] = Field(..., alias="mimeClass")
    content_type: Optional[str] = Field(..., alias="contentType")


class Page(ModuleItemInterface):
    m_id: str = Field(..., alias="_id")
    title: Optional[str] = Field(..., alias="title")
    updated_at: Optional[datetime] = Field(..., alias="updatedAt")
    created_at: Optional[datetime] = Field(..., alias="createdAt")


class ModuleItem(ConfiguredBaseModel):
    content: Optional[Union[File, Page, ModuleItemInterface]] = Field(..., alias="content")


class Module(ConfiguredBaseModel):
    name: Optional[str] = Field(..., alias="name")
    q_id: str = Field(..., alias="id")
    module_items: Optional[list[ModuleItem]] = Field(..., alias="moduleItems")


class ModuleConnection(ConfiguredBaseModel):
    nodes: Optional[list[Optional[Module]]] = Field(..., alias="nodes")


class CanvasCourseData(ConfiguredBaseModel):
    m_id: str = Field(..., alias="_id")
    name: str = Field(..., alias="name")
    course_nickname: Optional[str] = Field(..., alias="courseNickname")
    term: Optional[Term] = Field(..., alias="term")
    assignments_connection: Optional[AssignmentConnection] = Field(None, alias="assignmentsConnection")
    modules_connection: Optional[ModuleConnection] = Field(None, alias="modulesConnection")



from datetime import datetime

import qdarktheme

from qcanvas.util.app_settings import settings


def apply_selected_theme() -> None:
    """
    Applies the selected theme from the app's settings
    """
    if settings.theme != "native":
        qdarktheme.setup_theme(
            settings.theme,
            custom_colors=_get_colours()
        )


red_theme = {
    "primary": "e21d31",
    "[light]": {"foreground": "480910", "background": "fcf8f8"},
    "[dark]": {"foreground": "fbdfe2", "background": "231f1f"}
}


def _get_colours() -> dict:
    now = datetime.now()

    if now.year >= 2025:
        print("I certainly hope not")

    if now.month == 3 and now.day == 17:
        # And this is on the weekend...
        return {"primary": "08ff00"}
    elif now.month == 2 and now.day == 14:
        print("Why are you looking at canvas? Don't you have something better to do?")

        # Nobody will ever see this because uni starts around the 20th
        # Too bad, I kinda liked the theme
        return red_theme
    elif now.month == 8 and now.day == 20:
        # Some random day... I just wanted to see the red theme
        return red_theme
    else:
        return {"primary": "FF804F"}



import asyncio
import json
import os
from importlib.metadata import version

import httpx
from packaging.version import Version

from qcanvas.util.constants import package_name

# When true, signals that the program should be restarted when it closes next
restart_flag = False


async def do_update() -> None:
    """
    Updates the qcanvas package and sets the restart flag.
    The restart flag is passed back to the launcher script as a return code.
    """
    global restart_flag
    await asyncio.to_thread(os.system, f"pip install --upgrade {package_name}")
    restart_flag = True


async def get_newer_version() -> tuple[Version | None, Version | None] | None:
    """
    Check for a newer version of qcanvas
    Returns
    -------
    tuple
        A tuple, where the first item is the latest version and the second item is the installed version. If the installed
        version is up-to-date, then the first item is None
    """
    latest_version = await get_latest_version()
    installed_version = Version(version(package_name))

    print(f"latest = {latest_version}, installed = {installed_version}")

    if installed_version < latest_version:
        return latest_version, installed_version
    else:
        return None, installed_version


async def get_latest_version() -> Version:
    """
    Retrieves the latest version of the package from pypi
    Returns
    -------
    Version
        The latest version of qcanvas on pypi
    """
    async with httpx.AsyncClient() as client:
        data = json.loads((await client.get(f"https://pypi.org/pypi/{package_name}/json")).text)
        return Version(data["info"]["version"])



import os.path
import pathlib
import platform
import re
from datetime import datetime
from enum import Enum
from typing import List, Optional, MutableSequence

from sqlalchemy import ForeignKey, Text
from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase, mapped_column, Mapped, relationship, MappedAsDataclass

import qcanvas.util.tree_util as tree


class PageLike:
    @property
    def content(self) -> str | None:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        raise NotImplementedError()

    @property
    def id(self) -> str:
        raise NotImplementedError()

    @property
    def resources(self) -> MutableSequence["Resource"]:
        raise NotImplementedError()

    @property
    def course_id(self) -> str:
        raise NotImplementedError()

    @course_id.setter
    def course_id(self, value: str):
        raise NotImplementedError()

    @property
    def updated_at(self) -> datetime:
        raise NotImplementedError()


class Base(AsyncAttrs, DeclarativeBase):
    pass


def default_assignment_module(module: "Module") -> bool:
    result = module.name.lower() in ["assessments", "assessment"]

    return result


class GroupByPreference(Enum):
    GROUP_BY_PAGES = 0
    GROUP_BY_MODULES = 1


class CoursePreferences(Base):
    __tablename__ = "preferences"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    local_name: Mapped[Optional[str]]
    files_group_by_preference: Mapped["GroupByPreference"] = mapped_column(default=GroupByPreference.GROUP_BY_PAGES)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="preferences")


class Course(MappedAsDataclass, Base, init=False):
    __tablename__ = "courses"

    id: Mapped[str] = mapped_column(primary_key=True)

    term_id: Mapped[str] = mapped_column(ForeignKey("terms.id"))
    term: Mapped["Term"] = relationship(back_populates="courses")

    preferences: Mapped["CoursePreferences"] = relationship(back_populates="course")

    name: Mapped[str]

    modules: Mapped[List["Module"]] = relationship(back_populates="course")
    module_items: Mapped[List["ModuleItem"]] = relationship(back_populates="course")
    assignments: Mapped[List["Assignment"]] = relationship(back_populates="course")
    resources: Mapped[List["Resource"]] = relationship(back_populates="course")


class Term(MappedAsDataclass, Base):
    """
    A term object.
    Each course belongs to a term.

    Attributes
    ----------
    name: str
        Name of the term
    start_at: Optional[datetime]
        Date the term starts
        Will be null for the 'default term'.
    end_at: Optional[datetime]
        Date the term ends.
        Will be null for the 'default term'.
    courses: list[Course]
        *Not required in constructor.*
        List of courses that belong to this term.
    """

    __tablename__ = "terms"
    id: Mapped[str] = mapped_column(primary_key=True)

    end_at: Mapped[Optional[datetime]]
    start_at: Mapped[Optional[datetime]]
    name: Mapped[str]

    courses: Mapped[List["Course"]] = relationship(back_populates="term", cascade="all, delete")

    def __init__(self, id: str, name: str, end_at: Optional[datetime], start_at: Optional[datetime]):
        super().__init__()
        self.id = id
        self.name = name
        self.end_at = end_at
        self.start_at = start_at

    def __hash__(self):
        return hash(self.id) ^ hash(self.end_at) ^ hash(self.start_at) ^ hash(self.name)


class Module(MappedAsDataclass, Base, tree.HasText, init=False):
    __tablename__ = "modules"
    id: Mapped[str] = mapped_column(primary_key=True)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="modules")

    name: Mapped[str]

    items: Mapped[List["ModuleItem"]] = relationship(back_populates="module", order_by="ModuleItem.position")

    @property
    def text(self) -> str:
        return self.name


class ResourceToModuleItemAssociation(MappedAsDataclass, Base):
    __tablename__ = "resource_to_moduleitem"
    module_item_id: Mapped[str] = mapped_column(ForeignKey("module_items.id"), primary_key=True)
    resource_id: Mapped[str] = mapped_column(ForeignKey("resources.id"), primary_key=True)


class ResourceToAssignmentAssociation(MappedAsDataclass, Base):
    __tablename__ = "resource_to_assignment"
    assignment_id: Mapped[str] = mapped_column(ForeignKey("assignments.id"), primary_key=True)
    resource_id: Mapped[str] = mapped_column(ForeignKey("resources.id"), primary_key=True)


class ResourceState(Enum):
    NOT_DOWNLOADED = 0
    DOWNLOADED = 1
    FAILED = 2

    @staticmethod
    def human_readable(value: "ResourceState"):
        match value:
            case ResourceState.NOT_DOWNLOADED:
                return "Not downloaded"
            case ResourceState.DOWNLOADED:
                return "Downloaded"
            case ResourceState.FAILED:
                return "Failed"

        raise ValueError(value)


class Resource(MappedAsDataclass, Base, tree.HasText):
    __tablename__ = "resources"
    id: Mapped[str] = mapped_column(primary_key=True)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="resources")

    url: Mapped[str]
    file_name: Mapped[str]  # Human-readable name
    file_size: Mapped[int]
    state: Mapped[ResourceState]
    fail_message: Mapped[Optional[str]]
    date_discovered: Mapped[datetime]

    module_items: Mapped[List["ModuleItem"]] = relationship(secondary=ResourceToModuleItemAssociation.__table__,
                                                            back_populates="resources", order_by="ModuleItem.position")
    assignments: Mapped[List["Assignment"]] = relationship(secondary=ResourceToAssignmentAssociation.__table__,
                                                           back_populates="resources")

    def __init__(self, id: str, url: str, file_name: str, file_size: int,
                 date_discovered: datetime = datetime.now(),
                 fail_message: Optional[str] = None, state=ResourceState.NOT_DOWNLOADED):
        super().__init__()
        self.id = id
        self.url = url
        self.file_name = file_name
        self.file_size = file_size
        self.fail_message = fail_message
        self.state = state
        self.date_discovered = date_discovered

    def __eq__(self, __value):
        if isinstance(__value, ModulePage):
            return self.id == __value.id
        else:
            return super().__eq__(__value)

    @property
    def text(self) -> str:
        return self.file_name

    @property
    def download_location(self) -> pathlib.Path:
        file_id: str = self.id

        # Colon is illegal in microsoft windows file names
        if platform.system() == "Windows":
            file_id = file_id.replace(':', '$')

        file_name, file_extension = os.path.splitext(self.file_name)

        return pathlib.Path("download", self._sanitise_course_name(self.course.name),
                            f"{file_name} [{file_id}]{file_extension}")

    @staticmethod
    def _sanitise_course_name(name: str) -> str:
        return re.sub("[/\\\\<>:\"?|*]", "_", name)


class ModuleItem(MappedAsDataclass, Base, tree.HasText):
    __tablename__ = "module_items"

    id: Mapped[str] = mapped_column(primary_key=True)

    module_id: Mapped[str] = mapped_column(ForeignKey("modules.id"))
    module: Mapped["Module"] = relationship(back_populates="items")

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="module_items")

    created_at: Mapped[datetime]
    updated_at: Mapped[datetime]
    name: Mapped[str]
    position: Mapped[int]
    type: Mapped[str]

    resources: Mapped[List["Resource"]] = relationship(secondary=ResourceToModuleItemAssociation.__table__,
                                                       back_populates="module_items")

    __mapper_args__ = {
        "polymorphic_identity": "module_item",
        "polymorphic_on": "type",
    }

    def __init__(self, id: str, created_at: datetime, updated_at: datetime, name: str):
        super().__init__()

        self.id = id
        self.created_at = created_at
        self.updated_at = updated_at
        self.name = name

    @property
    def text(self) -> str:
        return self.name


class ModuleFile(ModuleItem):
    __tablename__ = "module_files"

    id: Mapped[str] = mapped_column(ForeignKey("module_items.id"), primary_key=True)

    __mapper_args__ = {
        "polymorphic_identity": "module_file",
    }

    def __init__(self, id: str, created_at: datetime, updated_at: datetime, name: str):
        super().__init__(id, created_at, updated_at, name)

        self.id = id


class ModulePage(ModuleItem, PageLike):
    __tablename__ = "module_pages"

    id: Mapped[str] = mapped_column(ForeignKey("module_items.id"), primary_key=True)

    content: Mapped[str] = mapped_column()

    __mapper_args__ = {
        "polymorphic_identity": "module_page",
    }

    def __init__(self, id: str, created_at: datetime, updated_at: datetime, name: str, content: str = "Not loaded"):
        super().__init__(id, created_at, updated_at, name)

        self.id = id
        self.content = content

    def __eq__(self, __value):
        if isinstance(__value, ModulePage):
            return self.id == __value.id
        else:
            return super().__eq__(__value)


class Assignment(MappedAsDataclass, Base, PageLike, tree.HasText, init=False):
    __tablename__ = "assignments"

    id: Mapped[str] = mapped_column(primary_key=True)

    course_id: Mapped[str] = mapped_column(ForeignKey("courses.id"))
    course: Mapped["Course"] = relationship(back_populates="assignments")

    name: Mapped[str]
    description: Mapped[Optional[str]] = mapped_column(Text, repr=False)
    due_at: Mapped[Optional[datetime]]
    created_at: Mapped[datetime]
    updated_at: Mapped[datetime]
    position: Mapped[int]

    resources: Mapped[List["Resource"]] = relationship(secondary=ResourceToAssignmentAssociation.__table__,
                                                       back_populates="assignments")

    @property
    def content(self) -> str:
        return self.description

    @property
    def text(self) -> str:
        return self.name



import asyncio
import json
import logging
from typing import BinaryIO, AsyncIterator, Any

import gql
import httpx
from gql.transport.exceptions import TransportQueryError
from httpx import URL, Response
from tenacity import retry, wait_exponential, wait_random, stop_after_attempt, wait_fixed, \
    retry_if_exception_type

import qcanvas.db as db
from qcanvas.net.canvas.legacy_canvas_types import LegacyFile, LegacyPage
from qcanvas.net.custom_httpx_async_transport import CustomHTTPXAsyncTransport
from qcanvas.net.self_authenticating import SelfAuthenticatingWithHttpClient, AuthenticationException


class RatelimitedException(Exception):
    def __init__(self):
        super().__init__("Canvas is ratelimiting me yay")


class UnauthenticatedException(Exception):
    def __init__(self):
        super().__init__("Trust me bro i'm authenticated")


def detect_unauthenticated_and_raise(response: Response) -> Response:
    """
    Detects if a response is redirecting us to the login page or is giving us a 401 (unauthorised)
    :param response: The response to check
    :return: True if reauthentication is needed
    """
    if response.url.path == "/login/canvas" or response.status_code == 401:
        raise UnauthenticatedException()

    return response


def detect_ratelimit_and_raise(response: Response) -> Response:
    # Who the FUCK decided to use 403 instead of 429?? With this stupid message??
    # Fuck you instructure, learn to code
    # And the newline at the end for some fucking reason is the cherry on top...
    if response.status_code == 403 and response.text == "403 Forbidden (Rate Limit Exceeded)\n":
        raise RatelimitedException()

    return response


class CanvasClient(SelfAuthenticatingWithHttpClient):
    _logger = logging.getLogger("canvas_client")
    _net_op_sem = asyncio.Semaphore(20)

    @staticmethod
    async def verify_config(canvas_url: str, api_key: str) -> bool:
        """
        Makes a request to canvas to verify that the url and key are correct.
        :param canvas_url: The canvas url to verify
        :param api_key: The api key to verify
        :return: True if everything looks ok, false if the api key or url is wrong
        """
        client = httpx.AsyncClient()
        # Make a request to an endpoint that returns very little/no data (for students at least) to check if everything
        # is working
        response = await client.get(url=URL(canvas_url).join("api/v1/accounts"),
                                    headers={"Authorization": f"Bearer {api_key}"}, timeout=10)

        return response.is_success

    def __init__(self, canvas_url: URL, api_key: str, client: httpx.AsyncClient | None = None):
        super().__init__(client=client or httpx.AsyncClient(timeout=60))
        self.api_key = api_key
        self.canvas_url = canvas_url

    def get_headers(self) -> dict[str, dict]:
        return {"headers": {"Authorization": f"Bearer {self.api_key}"}}

    @retry(
        wait=wait_exponential(exp_base=1.2, max=10) + wait_random(0, 1),
        retry=retry_if_exception_type(RatelimitedException),
        stop=stop_after_attempt(8)
    )
    async def get_courses(self):
        with self._net_op_sem:
            return await self.client.get(self.canvas_url.join("api/v1/courses"))

    @retry(
        wait=wait_exponential(exp_base=1.2, max=10) + wait_random(0, 1),
        retry=retry_if_exception_type(RatelimitedException),
        stop=stop_after_attempt(8)
    )
    async def get_page(self, page_id: str | int, course_id: str | int) -> LegacyPage:
        async with self._net_op_sem:
            response = detect_ratelimit_and_raise(
                await self.client.get(self.canvas_url.join(f"api/v1/courses/{course_id}/pages/{page_id}"),
                                      **self.get_headers()))

            return LegacyPage.from_dict(json.loads(response.text))

    @retry(
        wait=wait_exponential(exp_base=1.2, max=10) + wait_random(0, 1),
        retry=retry_if_exception_type(RatelimitedException),
        stop=stop_after_attempt(8)
    )
    async def get_file(self, file_id: str | int, course_id: str | int) -> LegacyFile:
        async with self._net_op_sem:
            response = detect_ratelimit_and_raise(
                await self.client.get(self.canvas_url.join(f"api/v1/courses/{course_id}/files/{file_id}"),
                                      **self.get_headers()))

            response.raise_for_status()

            return LegacyFile.from_dict(json.loads(response.text))

    @retry(
        wait=wait_exponential(exp_base=1.2, max=10) + wait_random(0, 1),
        retry=retry_if_exception_type(RatelimitedException),
        stop=stop_after_attempt(8)
    )
    async def get_file_from_endpoint(self, endpoint_url: str) -> LegacyFile:
        async with self._net_op_sem:
            response = detect_ratelimit_and_raise(await self.client.get(endpoint_url, **self.get_headers()))

            response.raise_for_status()

            return LegacyFile.from_dict(json.loads(response.text))

    @retry(
        wait=wait_exponential(exp_base=1.2, max=10) + wait_random(0, 1),
        retry=retry_if_exception_type(RatelimitedException),
        stop=stop_after_attempt(8)
    )
    async def get_temp_session_link(self) -> str | None:
        """
        Gets the link which will authenticate a browser/open canvas using a 'legacy session token'
        Returns
        -------
        str | None
            The url as a string if the request succeeded, None if it didn't
        """

        token_response = await self.client.get(self.canvas_url.join("login/session_token"), **self.get_headers())

        if token_response.is_success:
            return json.loads(token_response.text)["session_url"]
        else:
            return None

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_fixed(5) + wait_random(0, 1),
        retry=retry_if_exception_type(TransportQueryError)
    )
    async def do_graphql_query(self, query: gql.client.DocumentNode, **kwargs) -> dict[str, Any]:
        async with self._net_op_sem:
            gql_transport = CustomHTTPXAsyncTransport(self.client, self.canvas_url.join("api/graphql"),
                                                      **self.get_headers())
            gql_client = gql.Client(transport=gql_transport, execute_timeout=60)

            return await gql_client.execute_async(query, variable_values=kwargs)

    async def download_file(self, resource: db.Resource, download_destination: BinaryIO) -> AsyncIterator[int]:
        yield 0
        retries = 0

        while retries < self.max_retries:
            async with self.client.stream(method='GET', url=resource.url, cookies=self.client.cookies,
                                          follow_redirects=True) as resp:
                if await self.reauthenticate_if_needed(resp):
                    retries += 1
                    self._logger.warning("Retrying download of %s", resource.url)
                    continue

                async for chunk in resp.aiter_bytes():
                    download_destination.write(chunk)
                    yield resp.num_bytes_downloaded

                return

        self._logger.warning("Gave up download of %s", resource.url)
        raise

    async def do_request_and_retry_if_unauthenticated(self, url: URL, method: str, **kwargs) -> httpx.Response:
        # Add API token to the request
        return await super().do_request_and_retry_if_unauthenticated(url, method, **kwargs, **self.get_headers())

    def detect_authentication_needed(self, response: Response):
        # Canvas will silently redirect to the login page or give a 401 if we are not authenticated
        return response.url.path == "/login/canvas" or response.status_code == 401

    async def _authenticate(self):
        token_response = await self.client.get(self.canvas_url.join("login/session_token"), **self.get_headers())

        if token_response.is_success:
            session_url = json.loads(token_response.text)["session_url"]
            self._logger.debug("Got token response for authentication")

            if session_url is not None:
                # Headers are not needed here
                req = await self.client.get(session_url)
                if req.status_code != 302:
                    self._logger.error("Error when activating session from request")
                    raise AuthenticationException("Authentication was not successful")
            else:
                raise AuthenticationException("Token response body was malformed")
        else:
            self._logger.error("Authentication failed, API key may be invalid")
            raise AuthenticationException("Authentication failed, check your API key")



from typing import Sequence

from PySide6.QtWidgets import *  # QWidget, QTreeView, QGroupBox, QBoxLayout, QHeaderView, QHBoxLayout, QComboBox, QLabel

import qcanvas.db as db
from qcanvas.ui.viewer.file_list import FileList
from qcanvas.util.constants import default_assignments_module_names
from qcanvas.util.download_pool import DownloadPool


class FileColumn(QGroupBox):
    def __init__(self, column_name, download_pool: DownloadPool):
        super().__init__(title=column_name)

        self.tree = FileList(download_pool)
        self.setLayout(QBoxLayout(QBoxLayout.Direction.TopToBottom))
        self.layout().addWidget(self.tree)

    def load_items(self, items: Sequence[db.ModuleItem | db.Module]):
        self.tree.load_items(items)

    def clear(self):
        self.tree.clear()


class FileViewTab(QWidget):

    def __init__(self, download_pool: DownloadPool):
        super().__init__()

        self.files_column = FileColumn("Files", download_pool)
        self.assignment_files_column = FileColumn("Assignment files", download_pool)

        layout = QHBoxLayout()
        layout.addWidget(self.files_column)
        layout.addWidget(self.assignment_files_column)

        self.setLayout(layout)

    def load_course_files(self, course: db.Course):
        module_items: list[db.ModuleItem] = []
        assignment_items: list[db.ModuleItem] = []

        for module_item in course.module_items:
            if module_item.module.name.lower() in default_assignments_module_names:
                assignment_items.append(module_item)
            else:
                module_items.append(module_item)

        if course.preferences.files_group_by_preference == db.GroupByPreference.GROUP_BY_MODULES:
            exclude_assignments_module = list(
                filter(lambda x: x.name.lower() not in default_assignments_module_names, course.modules))

            self.files_column.load_items(exclude_assignments_module)
        else:
            self.files_column.load_items(module_items)

        self.assignment_files_column.load_items(assignment_items + course.assignments)

    def clear(self):
        self.files_column.clear()
        self.assignment_files_column.clear()



from typing import Any, TypeVar, Type, cast

T = TypeVar("T")


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class LegacyPage:
    published: bool
    hide_from_students: bool
    locked_for_user: bool
    body: str | None

    def __init__(self, published: bool, hide_from_students: bool, locked_for_user: bool, body: str | None) -> None:
        self.published = published
        self.hide_from_students = hide_from_students
        self.locked_for_user = locked_for_user
        self.body = body

    @staticmethod
    def from_dict(obj: Any) -> 'LegacyPage':
        assert isinstance(obj, dict)
        published = from_bool(obj.get("published"))
        hide_from_students = from_bool(obj.get("hide_from_students"))
        locked_for_user = from_bool(obj.get("locked_for_user"))
        body = None if locked_for_user else from_str(obj.get("body"))
        return LegacyPage(published, hide_from_students, locked_for_user, body)

    def to_dict(self) -> dict:
        result: dict = {}
        result["published"] = from_bool(self.published)
        result["hide_from_students"] = from_bool(self.hide_from_students)
        result["locked_for_user"] = from_bool(self.locked_for_user)
        result["body"] = from_str(self.body)
        return result


class LegacyFile:
    id: int
    uuid: str
    display_name: str
    filename: str
    url: str
    size: int
    locked: bool
    hidden: bool
    hidden_for_user: bool
    locked_for_user: bool

    def __init__(self, id: int, uuid: str, display_name: str, filename: str, url: str, size: int, locked: bool,
                 hidden: bool, hidden_for_user: bool, locked_for_user: bool) -> None:
        self.id = id
        self.uuid = uuid
        self.display_name = display_name
        self.filename = filename
        self.url = url
        self.size = size
        self.locked = locked
        self.hidden = hidden
        self.hidden_for_user = hidden_for_user
        self.locked_for_user = locked_for_user

    @staticmethod
    def from_dict(obj: Any) -> 'LegacyFile':
        assert isinstance(obj, dict)
        id = from_int(obj.get("id"))
        uuid = from_str(obj.get("uuid"))
        display_name = from_str(obj.get("display_name"))
        filename = from_str(obj.get("filename"))
        url = from_str(obj.get("url"))
        size = from_int(obj.get("size"))
        locked = from_bool(obj.get("locked"))
        hidden = from_bool(obj.get("hidden"))
        hidden_for_user = from_bool(obj.get("hidden_for_user"))
        locked_for_user = from_bool(obj.get("locked_for_user"))
        return LegacyFile(id, uuid, display_name, filename, url, size, locked, hidden, hidden_for_user, locked_for_user)

    def to_dict(self) -> dict:
        result: dict = {}
        result["id"] = from_int(self.id)
        result["uuid"] = from_str(self.uuid)
        result["display_name"] = from_str(self.display_name)
        result["filename"] = from_str(self.filename)
        result["url"] = from_str(self.url)
        result["size"] = from_int(self.size)
        result["locked"] = from_bool(self.locked)
        result["hidden"] = from_bool(self.hidden)
        result["hidden_for_user"] = from_bool(self.hidden_for_user)
        result["locked_for_user"] = from_bool(self.locked_for_user)
        return result


def legacy_page_from_dict(s: Any) -> LegacyPage:
    return LegacyPage.from_dict(s)


def legacy_page_to_dict(x: LegacyPage) -> Any:
    return to_class(LegacyPage, x)


def legacy_file_from_dict(s: Any) -> LegacyFile:
    return LegacyFile.from_dict(s)


def legacy_file_to_dict(x: LegacyFile) -> Any:
    return to_class(LegacyFile, x)



from typing import Sequence, Any


class HasColumnData:
    def get_column_data(self, column: int, role: int) -> str | None:
        raise NotImplementedError()


class HasText:
    @property
    def text(self) -> str:
        raise NotImplementedError()


class HasParent:
    @property
    def parent(self) -> Any:
        raise NotImplementedError()

    @property
    def index_of_self(self) -> int:
        raise NotImplementedError()


class HasChildren:
    @property
    def collapsed(self) -> bool:
        raise NotImplementedError()

    @collapsed.setter
    def collapsed(self, value: bool):
        raise NotImplementedError()

    @property
    def children(self) -> Sequence[HasColumnData]:
        raise NotImplementedError()



from PySide6.QtGui import QActionGroup
from PySide6.QtWidgets import QMenu, QWidget

from qcanvas.util.app_settings import settings
from qcanvas.util.helpers import theme_helper
from qcanvas.util.helpers.qaction_helper import create_qaction


def change_theme(theme_name: str):
    settings.theme = theme_name
    theme_helper.apply_selected_theme()


class ThemeSelectionMenu(QMenu):
    def __init__(self, parent: QWidget | None = None):
        super().__init__("Theme", parent)

        action_group = QActionGroup(self)

        light_theme = self._create_action("Light", "light")
        dark_theme = self._create_action("Dark", "dark")
        auto_theme = self._create_action("Auto (YMMV)", "auto")
        native_theme = self._create_action("Native (requires restart)", "native")

        actions = [light_theme, dark_theme, auto_theme, native_theme]

        self.addActions(actions)

        for theme in actions:
            action_group.addAction(theme)

    def _create_action(self, text: str, theme_name: str):
        return create_qaction(
            name=text,
            parent=self,
            triggered=lambda: change_theme(theme_name),
            checkable=True,
            checked=settings.theme == theme_name
        )



from typing import TypeVar, Generic

from PySide6.QtCore import QSettings
from packaging.version import Version

default_theme = "light"


def ensure_theme_is_valid(theme: str) -> str:
    """
    Ensures that a theme name is valid.
    If it is invalid, the default theme ("light") is returned
    Parameters
    ----------
    theme
        The theme name
    Returns
    -------
    str
        A valid theme name
    """
    if theme not in ["auto", "light", "dark", "native"]:
        return default_theme
    else:
        return theme


T = TypeVar("T")


class MappedSetting(Generic[T]):
    """
    Acts as a proxy for a named value in a QSettings object.
    Stores the value in memory when initialised and updates it accordingly, to protect it from changes on disk.
    """

    def __init__(self, settings_object: QSettings, setting_name: str, default: T | None = None):
        self.settings_object = settings_object
        self.setting_name = setting_name
        self.value = self.settings_object.value(self.setting_name, default)

    def __get__(self, instance, owner):
        return self.value

    def __set__(self, instance, value) -> None:
        self.value = value
        self.settings_object.setValue(self.setting_name, value)


class ThemeSetting(MappedSetting):
    def __init__(self, settings_object: QSettings):
        super().__init__(settings_object, "theme", default_theme)

    def __get__(self, instance, owner):
        return ensure_theme_is_valid(super().__get__(instance, owner))

    def __set__(self, instance, value):
        super().__set__(instance, ensure_theme_is_valid(value))


class _AppSettings:
    """
    Attributes
    ----------
    settings : QSettings
        Primary settings map for client settings
    auxiliary : QSettings
        Secondary settings map for settings which aren't related to canvas/panopto client functionality
    ignored_update
        If there is an update available and the user chooses not to update, then that version will be stored in here and
        the user will not be asked to update to that version again
    geometry
        Used to restore the main window position when it is re-opened
    window_state
        Used to restore the main window position when it is re-opened
    theme
        The theme of the app
    canvas_url
        The canvas url
    api_key
        The api key for canvas
    """

    settings = QSettings("QCanvas", "client")
    auxiliary = QSettings("QCanvas", "ui")

    canvas_url: MappedSetting[str] = MappedSetting(settings, "canvas_url")
    panopto_url: MappedSetting[str] = MappedSetting(settings, "panopto_url")
    api_key: MappedSetting[str] = MappedSetting(settings, "api_key")

    ignored_update: MappedSetting[Version] = MappedSetting(auxiliary, "ignored_update")
    theme: ThemeSetting = ThemeSetting(auxiliary)
    geometry = MappedSetting(auxiliary, "geometry")
    window_state = MappedSetting(auxiliary, "window_state")


# Global _AppSettings instance
settings = _AppSettings()
